import sys
import gc
import copy
from typing import Union, List, Tuple, Dict, Callable, Optional, Any, Iterable, Iterator, Generator
import contextlib
from contextlib import contextmanager
import warnings
import math

import torch
from torch.utils.data import Dataset
import numpy as np
from tqdm.auto import tqdm

from . import indexing
from . import misc


############################################
############ VARIABLE HELPERS ##############
############################################

def show_all_tensors(max_depth=1, verbose=0) -> None:
    """
    Displays all tensors present in the provided dictionary.
    From: https://discuss.pytorch.org/t/how-to-debug-causes-of-gpu-memory-leaks/6741
    RH 2024
    """
    # prints currently alive Tensors and Variables
    import torch
    import gc
    def get_info(obj, depth=0):
        if depth > max_depth:
            return
        
        if check_conditions(obj):
            objs[id(obj)] = {'type': type(obj), 'size': obj.size(), 'mem': misc.estimate_array_size(obj)}
            print(type(obj), obj.size(), misc.estimate_array_size(obj)) if verbose else None
        elif isinstance(obj, dict):
            for key, value in obj.items():
                get_info(value, depth+1)
        elif hasattr(obj, '__dict__'):
            get_info(obj.__dict__, depth+1)
        elif hasattr(obj, '__slots__'):
            for item in obj.__slots__:
                get_info(getattr(obj, item), depth+1)
        elif isinstance(obj, (list, tuple)):
            for item in obj:
                get_info(item, depth+1)
        
    def check_conditions(obj):
        return torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data))

    objs = {}
    for obj in gc.get_objects():
        try:
            get_info(obj)
        except:
            pass

    ## Sort by size
    objs = {k: v for k, v in sorted(objs.items(), key=lambda item: item[1]['mem'], reverse=True)}
    return objs



@contextmanager
def temp_eval(module):
    """
    Temporarily sets the network to evaluation mode within a context manager.
    RH 2024

    Args:
        module (torch.nn.Module):
            The network to temporarily set to evaluation mode.

    Yields:
        (torch.nn.Module):
            The network temporarily set to evaluation mode.

    Demo:
        .. highlight:: python
        .. code-block:: python

            with temp_eval(model):
                y = model(x)
    """
    state_train = module.training
    module.eval()
    try:
        yield module
    finally:
        if state_train:
            module.train()


######################################
############ CUDA STUFF ##############
######################################

def show_torch_cuda_info() -> None:
    """
    Displays PyTorch, CUDA, and device information.

    This function prints Python, PyTorch, CUDA, cuDNN version details and lists
    available CUDA devices. Note that the CUDA version cannot be directly
    fetched using Python, you should use the command `nvcc --version` in your
    terminal or `!nvcc --version` in a Jupyter notebook.

    The function also utilizes the `nvidia-smi` command to retrieve detailed
    information about the CUDA devices, including the GPU index, name, driver
    version, total memory, used memory, and free memory.
    RH 2021
    """
    print('Python version:', sys.version)
    print('PyTorch version:', torch.__version__)
    print('CUDA version: cannot be directly found with python function. Use `nvcc --version` in terminal or `! nvcc --version in notebook')
    from subprocess import call
    # ! nvcc --version
    print('CUDNN version:', torch.backends.cudnn.version())
    print('Devices:')
    call(["nvidia-smi", "--format=csv", "--query-gpu=index,name,driver_version,memory.total,memory.used,memory.free"])
    print ('Available torch cuda devices ', torch.cuda.device_count())
    print ('Current torch cuda device ', torch.cuda.current_device())


def show_cuda_devices(
    verbose: bool = True
) -> None:
    """
    Displays available CUDA devices.
    RH 2023

    Args:
        verbose (bool): 
            If ``True``, the function prints detailed information about each 
            device.
    """
    devices = [torch.cuda.get_device_properties(ii) for ii in range(torch.cuda.device_count())]
    if verbose:
        for ii, device in enumerate(devices):
            print(f'GPU {ii}: {device.name}, {device.total_memory/1000000000} GB')

def delete_all_cuda_tensors(
    globals: dict,
) -> None:
    """
    Deletes all CUDA tensors from the provided global variable dictionary.

    Call this function with: delete_all_cuda_tensors(globals())
    RH 2021

    Args:
        globals (dict):
            Dictionary of global variables.
            To obtain this, call the built-in ``globals()`` function.
    """
    types = [type(ii[1]) for ii in globals.items()]
    keys = list(globals.keys())
    for ii, (i_type, i_key) in enumerate(zip(types, keys)):
        if i_type is torch.Tensor:
            if globals[i_key].device.type == 'cuda':
                print(f'deleting: {i_key}, size: {globals[i_key].element_size() * globals[i_key].nelement()/1000000} MB')
                del(globals[i_key])
    
    clear_cuda_cache()

def clear_cuda_cache() -> None:
    """
    Clears the CUDA cache to free up memory resources.

    This function forces the garbage collector to release unreferenced memory and
    then clears the CUDA cache. This process is repeated four times in an attempt
    to maximize memory release.

    Note that this function will only have an effect if your machine has a GPU and
    if CUDA is enabled.
    """
    torch.cuda.empty_cache()
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()
    torch.cuda.empty_cache()


def wrap_clear_cuda_cache(func: Callable, before=True, after=True) -> Callable:
    """
    Wraps a function with a call to clear the CUDA cache before and/or after the
    function call.

    Args:
        func (Callable):
            The function to wrap.
        before (bool):
            If ``True``, clears the CUDA cache before calling the function.
            (Default is ``True``)
        after (bool):
            If ``True``, clears the CUDA cache after calling the function.
            (Default is ``True``)

    Returns:
        (Callable):
            The wrapped function.

    Demo:
        .. code-block:: python
            
                @wrap_clear_cuda_cache
                def my_function():
                    pass
    """
    import functools
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if before:
            clear_cuda_cache()
        result = func(*args, **kwargs)
        if after:
            clear_cuda_cache()
        return result
    return wrapper


def set_device(
    use_GPU: bool = True, 
    device_num: int = 0, 
    device_types: List[str] = ['cuda', 'mps', 'xpu', 'cpu'],
    verbose: bool = True
) -> str:
    """
    Sets the device for PyTorch. If a GPU is available and **use_GPU** is
    ``True``, it will be set as the device. Otherwise, the CPU will be set as
    the device. 
    RH 2022

    Args:
        use_GPU (bool): 
            Determines if the GPU should be utilized: \n
            * ``True``: the function will attempt to use the GPU if a GPU is
              not available.
            * ``False``: the function will use the CPU. \n
            (Default is ``True``)
        device_num (int): 
            Specifies the index of the GPU to use. (Default is ``0``)
        device_types (List[str]):
            The types and order of devices to attempt to use. The first device
            type that is available will be used. Options are ``'cuda'``,
            ``'mps'``, ``'xpu'``, and ``'cpu'``.
        verbose (bool): 
            Determines whether to print the device information. \n
            * ``True``: the function will print out the device information.
            \n
            (Default is ``True``)

    Returns:
        (str): 
            device (str): 
                A string specifying the device, either *"cpu"* or
                *"cuda:<device_num>"*.
    """
    devices = list_available_devices()

    if not use_GPU:
        device = 'cpu'
    else:
        device = None
        for device_type in device_types:
            if len(devices[device_type]) > 0:
                device = devices[device_type][device_num]
                break

    if verbose:
        print(f'Using device: {device}')

    return device
    

def list_available_devices() -> dict:
    """
    Lists all available PyTorch devices on the system.
    RH 2024

    Returns:
        (dict): 
            A dictionary with device types as keys and lists of available devices as values.
    """
    devices = {}

    # Check for CPU devices
    if torch.cpu.is_available():
        devices['cpu'] = ['cpu']
    else:
        devices['cpu'] = []

    # Check for CUDA devices
    if torch.cuda.is_available():
        devices['cuda'] = [f'cuda:{i}' for i in range(torch.cuda.device_count())]
    else:
        devices['cuda'] = []

    # Check for MPS devices
    if torch.backends.mps.is_available():
        devices['mps'] = ['mps:0']
    else:
        devices['mps'] = []

    # Check for XPU devices
    if hasattr(torch, 'xpu'):
        if torch.xpu.is_available():
            devices['xpu'] = [f'xpu:{i}' for i in range(torch.xpu.device_count())]
        else:
            devices['xpu'] = []
    else:
        devices['xpu'] = []

    return devices


def initialize_torch_settings(
    benchmark: Optional[bool] = None,
    enable_cudnn: Optional[bool] = None,
    deterministic_cudnn: Optional[bool] = None,
    deterministic_torch: Optional[bool] = None,
    set_global_device: Optional[Union[str, torch.device]] = None,
    init_linalg_device: Optional[Union[str, torch.device]] = None,
) -> None:
    """
    Initalizes some CUDA libraries and sets some environment variables. \n
    RH 2024

    Args:
        benchmark (Optional[bool]):
            If ``True``, sets torch.backends.cudnn.benchmark to ``True``.\n 
            This results in the built-in cudnn auto-tuner to find the best
            algorithm for the hardware. Good for when input sizes are the same
            for each batch.
        enable_cudnn (Optional[bool]):
            If ``True``, sets torch.backends.cudnn.enabled to ``True``.\n
            This enables the cudnn library.
        deterministic_cudnn (Optional[bool]):
            If ``True``, sets torch.backends.cudnn.deterministic to ``True``.\n
            This makes cudnn deterministic. It may slow down operations.
        deterministic_torch (Optional[bool]):
            If ``True``, sets torch.set_deterministic to ``True``.\n
            This makes torch deterministic. It may slow down operations.
        set_global_device (bool):
            If ``False``, does not set the global device. If a string or torch.device,
            sets the global device to the specified device.
        init_linalg_device (str):
            The device to use for initializing the linalg library. Either a
            string or a torch.device. This is necessary to avoid a bug. Often
            solves the error: "RuntimeError: lazy wrapper should be called at
            most once". (Default is ``None``)
    """
    if benchmark is not None:
        torch.backends.cudnn.benchmark = benchmark
    if enable_cudnn:
        torch.backends.cudnn.enabled = enable_cudnn
    if deterministic_cudnn:
        torch.backends.cudnn.deterministic = False
    if deterministic_torch:
        torch.set_deterministic(False)
    if set_global_device is not None:
        torch.cuda.set_device(set_global_device)
    
    ## Initialize linalg libarary
    ## https://github.com/pytorch/pytorch/issues/90613
    if init_linalg_device is not None:
        if type(init_linalg_device) is str:
            init_linalg_device = torch.device(init_linalg_device)
        torch.inverse(torch.ones((1, 1), device=init_linalg_device))
        torch.linalg.qr(torch.as_tensor([[1.0, 2.0], [3.0, 4.0]], device=init_linalg_device))


def profiler_simple(
    path_save: str = 'trace.json',
    activities: List[str] = ['CPU', 'CUDA'],
    with_stack: bool = False,
    record_shapes: bool = True,
    profile_memory: bool = True,
    with_flops: bool = False,
    with_modules: bool = False,
):
    """
    Simple profiler for PyTorch. \n
    Makes a context manager that can be used to profile code. \n
    Upon exit, will save the trace to the specified path. \n
    Use Chrome's chrome://tracing/ to view the trace. \n
    RH 2024

    Args:
        path_save (str):
            The path to save the trace.
        activities (List[str]):
            The activities to profile. Options are ``'CPU'`` and ``'CUDA'``.
        with_stack (bool):
            If ``True``, records the stack, which are all the functions that
            were called. 
        record_shapes (bool):
            If ``True``, records shapes of the tensors.
        profile_memory (bool):
            If ``True``, profiles memory usage of the tensors.
        with_flops (bool):
            If ``True``, records flops (floating point operations).
        with_modules (bool):
            If ``True``, records modules that were called.

    Returns:
        (contextmanager):
            A context manager that can be used to profile code.

    Demo:
        .. highlight:: python
        .. code-block:: python

            with profiler_simple():
                y = model(x)

            >> Then open google chrome and go to the url chrome://tracing/ and
            load the trace file.
    """
    from torch.profiler import profile, record_function, ProfilerActivity
    from contextlib import contextmanager
    
    activities_dict = {
        'CPU': ProfilerActivity.CPU,
        'CUDA': ProfilerActivity.CUDA,
    }
    activities = [activities_dict[act.upper()] for act in activities]

    @contextmanager
    def simple_profiler(path_save: str = 'trace.json'):
        with profile(
            activities=activities,
            record_shapes=record_shapes,
            profile_memory=profile_memory,
            with_stack=with_stack,
            with_flops=with_flops,
            with_modules=with_modules,
        ) as p:
            with record_function("model_inference"):
                yield
        p.export_chrome_trace(path_save)

    return simple_profiler(path_save=path_save)



######################################
############ DATA HELPERS ############
######################################


class Basic_dataset(Dataset):
    """
    Creates a basic PyTorch dataset. 
    RH 2021

    Args:
        X (torch.Tensor or np.ndarray): 
            The data from which to create the dataset. 
        device (str): 
            The device where the tensors will be stored. 
            (Default is ``'cpu'``)
        dtype (torch.dtype): 
            The data type to use for the tensor. 
            (Default is ``torch.float32``)

    Attributes:
        X (torch.Tensor):
            The data from which the dataset is created. The first (0th)
            dimension will be subsampled from.
        n_samples (int):
            The number of samples in the dataset.

    Example:
        .. highlight:: python
        .. code-block:: python

            ds = Basic_dataset(X, device='cuda:0')
            dl = DataLoader(ds, batch_size=32, shuffle=True)
    """
    def __init__(
        self,
        X: Union[torch.Tensor, np.ndarray],
        device: str = 'cpu',
        dtype: torch.dtype = torch.float32,
    ):
        """
        Initializes the Basic_dataset with the provided data, device, and dtype.
        """
        self.X = torch.as_tensor(X, dtype=dtype, device=device) # first (0th) dim will be subsampled from
        self.n_samples = self.X.shape[0]
        
    def __len__(self) -> int:
        """
        Returns the number of samples in the dataset.

        Returns:
            (int):
                n_samples (int):
                    The number of samples in the dataset.
        """
        return self.n_samples
    
    def __getitem__(
        self,
        idx: int,
    ) -> Tuple[torch.Tensor, int]:
        """
        Returns a single sample and its index from the dataset.

        Args:
            idx (int):
                The index of the sample to return.

        Returns:
            (tuple): tuple containing:
                sample (torch.Tensor):
                    The requested sample from the dataset.
                idx (int):
                    The index of the requested sample.
        """
        return self.X[idx]
    

class Dataset_numpy(Dataset):
    """
    Creates a PyTorch dataset from a numpy array. 
    RH 2024

    Args:
        X (np.ndarray):
            The data from which to create the dataset.
        axis (int):
            The dimension along which to sample the data.
        device (str): 
            The device where the tensors will be stored. 
        dtype (torch.dtype): 
            The data type to use for the tensor.

    Attributes:
        X (np.ndarray or np.memmap):
            The data from the numpy file.
        n_samples (int):
            The number of samples in the dataset.

    Returns:
        (torch.utils.data.Dataset):
            A PyTorch dataset.
    """
    def __init__(
        self,
        X: Union[np.ndarray, np.memmap],
        axis: int = 0,
        device: str = 'cpu',
        dtype: torch.dtype = torch.float32,
    ):
        """
        Initializes the Dataset_NumpyFile with the provided parameters.
        """
        assert isinstance(X, (np.ndarray, np.memmap)), 'X must be a numpy array or memmap.'
        self.X = X
        self.n_samples = self.X.shape[axis]
        self.is_memmap = isinstance(self.X, np.memmap) 
        self.axis = axis
        self.device = device
        self.dtype = dtype

    def __len__(self) -> int:
        """
        Returns the number of samples in the dataset.

        Returns:
            (int):
                n_samples (int):
                    The number of samples in the dataset.
        """
        return self.n_samples
    
    def __getitem__(
        self,
        idx: int,
    ) -> Tuple[torch.Tensor, int]:
        """
        Returns a single sample and its index from the dataset.

        Args:
            idx (int):
                The index of the sample to return.

        Returns:
            sample (torch.Tensor):
                The requested sample from the dataset.
        """
        arr = np.take(self.X, idx, axis=self.axis)
        if self.is_memmap:
            arr = np.array(arr)
        return torch.as_tensor(arr, dtype=self.dtype, device=self.device)
    
    def close(self):
        """
        Closes the numpy file.
        """
        if self.is_memmap:
            self.X.close()


class Dataset_numpy_concatenated(torch.utils.data.Dataset):
    """
    Dataset class for loading slices from arrays within a multiple numpy
    arrays.\n
    Input is a list of numpy arrays with similar shapes.\n
    Output is a Dataset where the queried index pulls slices from the
    concatenated first dimension indices of all the input arrays.\n
    RH 2024

    Args:
        arrays (list):
            A list of numpy arrays with the following organization: \n
                * Each array may have different first dimension sizes, but all
                  other dimensions must be the same.
    """
    def __init__(self, arrays: List[np.ndarray], verbose: bool = True):
        super(Dataset_numpy_concatenated, self).__init__()

        self.arrays = arrays

        ## Check that all arrays have the same shape except for the first dimension
        shapes = [arr.shape for arr in arrays]
        check_shape = lambda shape1, shape2: shape1[1:] == shape2[1:] if len(shape1) > 1 else shape1[0] == shape2[0]
        assert all([check_shape(shape, shapes[0]) for shape in shapes]), "All arrays must have the same shape except for the first dimension."
        self.n_samples = sum([shape[0] for shape in shapes])
        self.shape = [self.n_samples] + list(shapes[0][1:])

        ## Create an index to field mapping
        ### Use a binary search to find the field for a given index using the cumsum of the first dimensions
        self.cumsum = np.cumsum([0] + [shape[0] for shape in shapes], dtype=np.int64)
        self.idx_to_idxArray = lambda idx: np.searchsorted(self.cumsum, idx, side='right') - 1

    def __len__(self):
        return self.n_samples

    def __getitem__(self, idx):
        if (idx < 0):
            idx = self.n_samples + idx
        elif (idx >= self.n_samples):
            raise IndexError(f"Index {idx} is out of bounds for dataset of length {self.n_samples}.")
        
        idx_array = self.idx_to_idxArray(idx)
        idx_withinArray = idx - self.cumsum[idx_array]
        sample = self.arrays[idx_array][idx_withinArray]

        ## Return as a cloned tensor
        # return torch.as_tensor(sample.copy())
        return torch.as_tensor(sample)


class BatchRandomSampler(torch.utils.data.Sampler):
    """
    Creates a sampler similar to torch.utils.data.BatchSampler, but allows for
    randomized sampling of the batches. Wraps indexing.make_batches.
    RH 2024

    Args:
        len_dataset(int):
            The number of samples in the dataset.
        batch_size (int):
            Size of each batch.\n
            if None, then ``batch_size`` based on ``num_batches``.
        num_batches (int):
            Number of batches to make.\n
            if None, then ``num_batches`` based on ``batch_size``.
        min_batch_size (int):
            Minimum size of each batch. Set to ``-1`` to equal ``batch_size``.
        randomize_batch_indices
            Whether to randomize the transition indices between batches.\n
            So ``[(1,2,3), (4,5,6),]`` can become ``[(1,), (2,3,4), (5,6),]``.
        shuffle_batch_order (bool):
            Whether to shuffle the order in which batches are yielded.
        shuffle_iterable_order (bool):
            Whether to shuffle the order of the contents of the iterable before
            batching.

    Returns:
        (torch.utils.data.Sampler):
            A sampler for the dataset.
    """
    def __init__(
        self,
        len_dataset: int,
        batch_size: Optional[int] = None,
        num_batches: Optional[int] = None,
        min_batch_size: int = 1,
        randomize_batch_indices: bool = False,
        shuffle_batch_order: bool = False,
        shuffle_iterable_order: bool = False,
    ):
        """
        Initializes the BatchRandomSampler with the provided parameters.
        """
        assert isinstance(len_dataset, int), 'len_dataset must be an integer.'
        self.len_dataset = len_dataset
        self.batch_size = batch_size
        self.num_batches = num_batches
        self.min_batch_size = min_batch_size
        self.randomize_batch_indices = randomize_batch_indices
        self.shuffle_batch_order = shuffle_batch_order
        self.shuffle_iterable_order = shuffle_iterable_order

    def __iter__(self):
        """
        Returns an iterator over the batches.
        """
        self.batches = indexing.make_batches(
            iterable=torch.arange(self.len_dataset),
            batch_size=self.batch_size,
            num_batches=self.num_batches,
            min_batch_size=self.min_batch_size,
            randomize_batch_indices=self.randomize_batch_indices,
            shuffle_batch_order=self.shuffle_batch_order,
            shuffle_iterable_order=self.shuffle_iterable_order,
        )
        return iter(self.batches)
    
    def __len__(self):
        """
        Returns the number of samples in the dataset.
        """
        return self.len_dataset


def map_batches_cuda(
    batches: Union[Iterable, Iterator, Generator],
    func: Callable,
    device_func: str = 'cuda:0',
    device_return: str = 'cpu',
    pin_memory: bool = True,
    non_blocking: bool = True,
    progress_bar: bool = False,
    len_batches: Optional[int] = None,
    verbose: bool = False,
):
    """
    Run batches through a user specified function on a cuda device using CUDA
    streams. This function is useful for running batches through a model in
    serial, but with asynchronous data transfers. This is similar to
    torch.utils.data.DataLoader, but is simpler and doesn't implement
    multiprocessing.
    RH 2024

    Args:
        batches (Union[Iterable, Iterator, Generator]):
            The batches to process. Each batch can either be a single tensor, a
            tuple containing tensors, or other.
        func (Callable):
            The function to run on the batches. The function must accept the
            items in the batches as arguments.
        device_func (str):
            The device to run the function on. Should typically be a CUDA device
            (e.g., 'cuda:0'). Can also be 'cpu' or other, but this will not
            utilize CUDA streams, memory pinning, or non-blocking transfers.
            (Default is 'cuda:0')
        device_return (str):
            The device to return the results on. For this function, it will
            typically be 'cpu'. (Default is 'cpu')
        pin_memory (bool):
            If ``True``, the function will pin the memory for the input tensors.
            Ideally, the original, non-batched, tensors should be pinned.
            However, this option will pin the memory for each batch. Warning,
            pinning memory can cause excess and persistent memory usage.
            (Default is ``True``)
        non_blocking (bool):
            If ``True``, the function will use non-blocking transfers to and
            from the device. Warning, pinning memory can cause excess and
            persistent memory usage. (Default is ``True``)
        progress_bar (bool):
            If ``True``, displays a progress bar. (Default is ``False``)
        len_batches Optional[int]:
            The length of the batches to be used as the total length of the
            progress bar. If not provided, will attempt to find len(batches).
            (Default is ``None``)
        verbose (bool):
            If ``True``, displays warnings. (Default is ``False``)

    Returns:
        (List):
            results (List):
                The results of the function run on the batches.

    Example:
        .. highlight:: python
        .. code-block:: python

            def func(x, y):
                a, b = model.forward(x, y)
                return a, b
            
            batches = ((x_i, y_i) for x_i, y_i in zip(X.pin_memory(), Y.pin_memory()))
        
            results = process_batches_cuda(
                batches,
                func,
                device_func='cuda:0',
                device_return='cpu',
                pin_memory=False,
                non_blocking=True,
                progress_bar=True,
                len_batches=int(math.ceil(len(X)/batch_size),
                verbose=False,
            )
    """
    results = []
    device_func = torch.device(device_func)

    assert device_func.type in ['cuda', 'cpu'], 'device must be either "cuda" or "cpu" type.'

    if len_batches is None:
        len_batches = len(batches) if hasattr(batches, '__len__') else None
    
    def send_to_device(batch, device, non_blocking, pin_memory):
        """
        Send a batch (either a single tensor, tuple containing tensors, or
        other) to a device.
        """
        if isinstance(batch, torch.Tensor):
            if pin_memory and batch.device.type != 'cuda':
                batch = batch.pin_memory()

            if batch.device != device:
                if non_blocking:
                    batch = batch.to(device, non_blocking=non_blocking)
                else:
                    batch = batch.to(device)

        elif isinstance(batch, tuple):
            batch = tuple([send_to_device(b, device, non_blocking, pin_memory) for b in batch])
        
        return batch

    if device_func.type == 'cuda':
        stream = torch.cuda.Stream(device=device_func)  ## create a stream
        for ii, batch in tqdm(enumerate(batches), total=len_batches, disable=not progress_bar):
            if not isinstance(batch, tuple):  ## Make batch a tuple
                batch = (batch,)
            with torch.cuda.stream(stream):  ## run on new stream
                batch = send_to_device(batch=batch, device=device_func, non_blocking=non_blocking, pin_memory=pin_memory)
                outs = func(*batch)  ## run func on batch
                outs = send_to_device(batch=outs, device=device_return, non_blocking=non_blocking, pin_memory=False)
                results.append(outs)  ## append to results
        torch.cuda.synchronize()  ## wait for all streams to finish
    else:
        warnings.warn('device_func is not cuda. No streams will be used.') if verbose else None
        for ii, batch in tqdm(enumerate(batches), total=len_batches, disable=not progress_bar):
            ## Make batch a tuple
            if not isinstance(batch, tuple):
                batch = (batch,)
            batch = send_to_device(batch=batch, device=device_func, non_blocking=non_blocking, pin_memory=pin_memory)
            outs = func(*batch)
            outs = send_to_device(batch=outs, device=device_return, non_blocking=non_blocking, pin_memory=False)
            results.append(outs)

    return results


##################################################################
############# STUFF PYTORCH SHOULD ALREADY HAVE ##################
##################################################################


numpy_to_torch_dtype_dict = {
    np.bool_      : torch.bool,
    np.uint8      : torch.uint8,
    np.int8       : torch.int8,
    np.int16      : torch.int16,
    np.int32      : torch.int32,
    np.int64      : torch.int64,
    np.float16    : torch.float16,
    np.float32    : torch.float32,
    np.float64    : torch.float64,
    np.complex64  : torch.complex64,
    np.complex128 : torch.complex128
}


@misc.wrapper_flexible_args(['dim', 'axis'])
@misc.wrapper_flexible_args(['keepdim', 'keepdims'])
def nanvar(
    arr: torch.Tensor, 
    dim: Optional[int] = None, 
    keepdim: bool = False, 
    ddof: int = 0, 
    ret_mean: bool = False
) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
    """
    Computes the variance of an array ignoring any NaN values.
    RH 2022

    Args:
        arr (torch.Tensor): 
            The input tensor from which to calculate the variance. 
        dim (Optional[int]): 
            The dimension along which to compute the variance. 
            If ``None``, computes variance over all dimensions. (Default is ``None``)
        keepdim (bool): 
            If ``True``, the output tensor is of the same size as input 
            except in the dimension(s) dim where it is of size 1. 
            Otherwise, dim is squeezed. (Default is ``False``)
        ddof (int): 
            Delta degrees of freedom. (Default is *0*)
        ret_mean (bool): 
            If ``True``, return mean along with variance. (Default is ``False``)

    Returns:
        (Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]):
            var (torch.Tensor):
                Variance of the input tensor ignoring any NaN values.
            avg (torch.Tensor):
                Mean of the input tensor ignoring any NaN values. This is only returned if ret_mean is ``True``.
    """
    nan_mask = torch.isnan(arr)
    arr_no_nan = arr.masked_fill(nan_mask, 0)
    cnt = torch.sum(torch.logical_not(nan_mask), dim=dim, keepdim=True)
    avg = torch.sum(arr_no_nan, dim=dim, keepdim=True) / cnt
    dev_sqrd = (arr_no_nan - avg)**2
    var = torch.sum(dev_sqrd, dim=dim, keepdim=True) / cnt
    dof = cnt - ddof
    out = var * cnt / dof

    out = squeeze_multiple_dims(out, dims=dim) if keepdim==False else out
    return out if ret_mean==False else (out, avg)

@misc.wrapper_flexible_args(['dim', 'axis'])
@misc.wrapper_flexible_args(['keepdim', 'keepdims'])
def nanstd(
    arr: torch.Tensor, 
    dim: Optional[int] = None, 
    keepdim: bool = False, 
    ddof: int = 0, 
    ret_mean: bool = False
) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
    """
    Computes the standard deviation of an array ignoring any NaN values. 
    RH 2022

    Args:
        arr (torch.Tensor): 
            The input tensor from which to calculate the standard deviation. 
        dim (Optional[int]): 
            The dimension along which to compute the standard deviation. If
            ``None``, computes standard deviation over all dimensions. (Default
            is ``None``)
        keepdim (bool): 
            If ``True``, the output tensor is of the same size as input except
            in the dimension(s) dim where it is of size 1. Otherwise, dim is
            squeezed. (Default is ``False``)
        ddof (int): 
            Delta degrees of freedom. (Default is *0*)
        ret_mean (bool): 
            If ``True``, return mean along with standard deviation. (Default is
            ``False``)

    Returns:
        (Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]):
            std (torch.Tensor):
                Standard deviation of the input tensor ignoring any NaN values.
            avg (torch.Tensor):
                Mean of the input tensor ignoring any NaN values. This is only
                returned if ret_mean is ``True``.
    """
    var = nanvar(arr, dim=dim, keepdim=keepdim, ddof=ddof, ret_mean=ret_mean)
    if ret_mean:
        var, avg = var
    std = torch.sqrt(var)
    return std if ret_mean==False else (std, avg)
        
@misc.wrapper_flexible_args(['dim', 'axis'])
@misc.wrapper_flexible_args(['keepdim', 'keepdims'])
def nanmax(
    arr: torch.Tensor, 
    dim: Optional[int] = None, 
    keepdim: bool = False
) -> torch.Tensor:
    """
    Computes the max of an array ignoring any NaN values.
    RH 2022

    Args:
        arr (torch.Tensor): 
            The input tensor from which to calculate the max value. 
        dim (Optional[int]): 
            The dimension along which to compute the max. If ``None``, computes
            max over all dimensions. (Default is ``None``)
        keepdim (bool): 
            If ``True``, the output tensor is of the same size as input except
            in the dimension(s) dim where it is of size 1. Otherwise, dim is
            squeezed. (Default is ``False``)

    Returns:
        (torch.Tensor):
            max_val (torch.Tensor):
                Maximum value of the input tensor ignoring any NaN values.
    """
    if dim is None:
        kwargs = {}
    else:
        kwargs = {
            'dim': dim,
            'keepdim': keepdim,
        }
    
    nan_mask = torch.isnan(arr)
    ## Get min value for the dtype. Use -inf if dtype is float, otherwise use min value for the dtype
    val_min = torch.as_tensor(float('-inf'), dtype=arr.dtype, device=arr.device) if arr.dtype in [torch.float32, torch.float64] else torch.iinfo(arr.dtype).min
    arr_no_nan = arr.masked_fill(nan_mask, val_min)
    return torch.max(arr_no_nan, **kwargs)

@misc.wrapper_flexible_args(['dim', 'axis'])
@misc.wrapper_flexible_args(['keepdim', 'keepdims'])
def nanmin(
    arr: torch.Tensor, 
    dim: Optional[int] = None, 
    keepdim: bool = False
) -> torch.Tensor:
    """
    Computes the min of an array ignoring any NaN values.
    RH 2022

    Args:
        arr (torch.Tensor): 
            The input tensor from which to calculate the min value. 
        dim (Optional[int]): 
            The dimension along which to compute the min. If ``None``, computes
            min over all dimensions. (Default is ``None``)
        keepdim (bool): 
            If ``True``, the output tensor is of the same size as input except
            in the dimension(s) dim where it is of size 1. Otherwise, dim is
            squeezed. (Default is ``False``)

    Returns:
        (torch.Tensor):
            min_val (torch.Tensor):
                Minimum value of the input tensor ignoring any NaN values.
    """
    if dim is None:
        kwargs = {}
    else:
        kwargs = {
            'dim': dim,
            'keepdim': keepdim,
        }
    
    nan_mask = torch.isnan(arr)
    ## Get max value for the dtype. Use inf if dtype is float, otherwise use max value for the dtype
    val_max = torch.as_tensor(float('inf'), dtype=arr.dtype, device=arr.device) if arr.dtype in [torch.float32, torch.float64] else torch.iinfo(arr.dtype).max
    arr_no_nan = arr.masked_fill(nan_mask, val_max)
    return torch.min(arr_no_nan, **kwargs)


def rolling_mean(tensor: torch.Tensor, dim: int) -> torch.Tensor:
    """
    Computes the running mean along a specified dimension using a rolling accumulation method
    (Welford's update for the mean).

    RH 2025

    Args:
        tensor (torch.Tensor):
            The input tensor on which the running mean is computed.
        dim (int):
            The dimension along which to compute the running mean.

    Returns:
        torch.Tensor:
            A tensor of the same shape as `tensor`, where each element along the specified dimension
            is the running mean of the elements from the start up to that index.
    """
    # Ensure the dimension is non-negative and valid.
    if dim < 0:
        dim += tensor.dim()
    if dim < 0 or dim >= tensor.dim():
        raise ValueError(f"Invalid dimension {dim} for tensor with {tensor.dim()} dimensions.")
    
    # Unbind the tensor along the given dimension to get a list of slices.
    dims_permute = list(range(tensor.dim()))
    ## remove dim from the list
    dims_permute.remove(dim)
    dims_permute = [dim] + dims_permute
        
    # Initialize an empty list to store the running means.
    current_mean = None
    
    # Iterate through each slice along the given dimension.
    # Use a counter starting at 1 since we divide by the count.
    for i, slice in enumerate(tensor.permute(dims_permute)):
        if current_mean is None:
            # For the first element, the running mean is the element itself.
            current_mean = slice
        else:
            # Update the running mean using:
            current_mean = current_mean + (slice - current_mean) / (i + 1)
    
    # Stack the list of running means back into a tensor along the specified dimension.
    return current_mean


def unravel_index(
    index: int, 
    shape: Tuple[int]
) -> List[int]:
    """
    Converts a flat index into a coordinate in a tensor of certain shape.
    RH 2022

    Args:
        index (int): 
            The flat index to be converted into a coordinate.
        shape (Tuple[int]): 
            The shape of the tensor in which the coordinate is calculated.

    Returns:
        (List[int]): 
            coord (List[int]): 
                The coordinate in the tensor.
    """
    out = []
    for dim in shape[::-1]:
        out.append(index % dim)
        index = index // dim
    return tuple(out[::-1])

@misc.wrapper_flexible_args(['dim', 'axis'])
def squeeze_multiple_dims(
    arr: torch.Tensor, 
    dims: Tuple[int, int] = (0, 1)
) -> torch.Tensor:
    """
    Squeezes multiple dimensions of a tensor.
    RH 2022

    Args:
        arr (torch.Tensor): 
            The input tensor to squeeze.
        dims (Tuple[int, int]): 
            The dimensions to squeeze. \n
            * dims[0]: Dimension to squeeze.
            * dims[1]: Dimension to squeeze. \n
            (Default is ``(0, 1)``)

    Returns:
        (torch.Tensor): 
            arr_squeezed (torch.Tensor):
                The squeezed tensor.
    """
    assert all([arr.shape[d] == 1 for d in dims])
    ## make custom slices to squeeze out the dims
    slices = [slice(None)] * arr.ndim
    for d in dims:
        slices[d] = 0
    return arr[tuple(slices)]


def multiply_elementwise_sparse_dense(
    s: torch.sparse.FloatTensor, 
    d: torch.Tensor
) -> torch.Tensor:
    """
    Multiplies a sparse tensor (s) with a dense tensor (d) elementwise.
    **Warning:** This creates an intermediate dense tensor with the same shape as s.
    RH 2022

    Args:
        s (torch.sparse.FloatTensor): 
            Sparse tensor to multiply.
        d (torch.Tensor): 
            Dense tensor to multiply.

    Returns:
        (torch.Tensor): 
            multiplied_tensor (torch.Tensor):
                Tensor resulting from the elementwise multiplication of `s` and
                `d`.
    """
    return torch.mul(d.expand(s.shape).sparse_mask(s.coalesce()), s)

def permute_sparse(
    input: torch.sparse.FloatTensor, 
    dims: List[int]
) -> torch.sparse.FloatTensor:
    """
    Permutes the dimensions of a sparse tensor.
    Adapted from https://github.com/pytorch/pytorch/issues/78422

    Args:
        input (torch.sparse.FloatTensor): 
            The input sparse tensor to permute.
        dims (List[int]): 
            The desired ordering of dimensions.

    Returns:
        (torch.sparse.FloatTensor): 
            permuted_tensor (torch.sparse.FloatTensor):
                The permuted sparse tensor.
    """
    dims = torch.LongTensor(dims)
    return torch.sparse_coo_tensor(indices=input._indices()[dims], values=input._values(), size=torch.Size(torch.tensor(input.size())[dims]))

def roll_sparse(
    X: torch.sparse.FloatTensor, 
    shifts: Union[int, Tuple[int]], 
    dims: Union[int, Tuple[int]]
) -> torch.sparse.FloatTensor:
    """
    Rolls a sparse tensor along the specified dimensions.

    Args:
        X (torch.sparse.FloatTensor): 
            Pytorch sparse tensor to roll.
        shifts (Union[int, Tuple[int]]): 
            Number of places by which elements are shifted.
        dims (Union[int, Tuple[int]]): 
            The dimensions along which the tensor is rolled.

    Returns:
        (torch.sparse.FloatTensor): 
            X_out (torch.sparse.FloatTensor): 
                The rolled sparse tensor.

    RH 2022
    """
    if type(shifts) is not tuple:
        shifts = (shifts,)
    if type(dims) is not tuple:
        dims = (dims,)

    X_out = copy.copy(X)
    for shift, dim in zip(shifts, dims):
        idx = X_out._indices()
        idx[dim] = (idx[dim] + shift) % X_out.shape[dim]
        X_out = torch.sparse_coo_tensor(indices=idx, values=X.coalesce().values(), size=X_out.shape)
    return X_out

def diag_sparse(
    x: torch.sparse.FloatTensor, 
    return_sparse_vals: bool = True
) -> torch.Tensor:
    """
    Gets the diagonal of a sparse tensor.

    Args:
        x (torch.sparse.FloatTensor): 
            Pytorch sparse tensor to extract the diagonal from.
        return_sparse_vals (bool): \n
            * If ``True``, returns 'sparse' (zeroed) values as well. 
            * If ``False``, returns only specified (non-sparsed out) values. 

    Returns:
        (torch.Tensor): 
            Diagonal values from the input sparse tensor.

    RH 2022
    """
    if return_sparse_vals is False:
        row, col = x.indices()
        values = x.values()
        return values[row == col]
    else:
        x_ts = indexing.torch_to_torchSparse(x)
        return x_ts.get_diag()

class Diag_sparse:
    """
    Gets the diagonal of a sparse tensor.
    The indices are kept as a property of the class for faster access.

    RH 2022

    Attributes:
        idx (torch.Tensor): 
            Indices of the diagonal elements in the sparse tensor.

    Args:
        x (torch.sparse.FloatTensor): 
            Sparse tensor to get diagonal of. For the __init__, 
            only the indices of the diagonal are kept as a property. 
            Subsequent calls of the class can use any x with the same indices.
    """
    def __init__(self, x):
        """
        Initializes the Diag_sparse class and stores the indices of the
        diagonal.
        """
        row, col = x.indices()
        self.idx = (row == col)
    def __call__(self, x: torch.sparse.FloatTensor) -> torch.Tensor:
        """
        Returns the diagonal elements from the sparse tensor using stored
        indices.

        Args:
            x (torch.sparse.FloatTensor): 
                Sparse tensor to get diagonal of.

        Returns:
            (torch.Tensor): 
                Diagonal values from the input sparse tensor.
        """
        values = x.values()
        return values[self.idx]


@misc.wrapper_flexible_args(['dim', 'axis'])
def zscore(
    X: torch.Tensor, 
    dim: Optional[int] = None, 
    ddof: int = 0, 
    nan_policy: str = 'propagate', 
) -> torch.Tensor:
    """
    Computes the z-score of a tensor.

    Args:
        X (torch.Tensor): 
            Tensor to compute z-score of.
        dim (Optional[int]): 
            Dimension to compute z-score over. 
            (Default is ``None``)
        ddof (int): 
            Means Delta Degrees of Freedom. 
            The divisor used in calculations is ``N - ddof``, 
            where ``N`` represents the number of elements. 
            (Default is *0*)
        nan_policy (str): 
            Defines how to handle when input contains nan. 
            The following options are available (default is ``'propagate'``): \n
            * ``'propagate'``: returns nan
            * ``'raise'``: throws an error
            * ``'omit'``: performs the calculations ignoring nan values \n
            (Default is ``'propagate'``)

    Returns:
        (torch.Tensor): 
            Z-scored tensor.

    RH 2022
    """
    assert dim is not None, 'Must specify dimension to compute z-score over.'

    if nan_policy == 'omit':
        std, mean = nanstd(X, dim=dim, ddof=ddof, keepdim=True, ret_mean=True)
    elif nan_policy == 'propagate':
        std, mean = torch.std_mean(X, dim=dim, keepdim=True, unbiased=ddof==1)
    else:
        raise ValueError('nan_policy must be "omit" or "propagate".')
    return (X - mean) / std


@misc.wrapper_flexible_args(['dim', 'axis'])
def slice_along_dim(
    X: torch.Tensor, 
    dim: int, 
    idx: Union[int, slice, List[int], torch.Tensor]
) -> torch.Tensor:
    """
    Slices a tensor along a specified dimension.
    RH 2022

    Args:
        X (torch.Tensor): 
            Tensor to slice.
        dim (int): 
            Dimension to slice along.
        idx Union[int, slice, List[int], torch.Tensor]:
            Index / slice / list of indices / tensor of indices to slice with.

    Returns:
        (torch.Tensor): 
            sliced_tensor (torch.Tensor):
                Sliced tensor.
    """
    slices = [slice(None)] * X.ndim
    slices[dim] = idx
    return X[tuple(slices)]


def orthogonal_procrustes(
    A: torch.Tensor,
    B: torch.Tensor,
    check_finite: bool = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Port of the scipy.linalg.orthogonal_procrustes function:
    https://github.com/scipy/scipy/blob/v1.13.0/scipy/linalg/_procrustes.py

    Computes the matrix solution of the orthogonal Procrustes problem.
    Given two matrices, A and B, find the orthogonal matrix that most closely
    maps A to B using the algorithm in [1].

    Args:
        A (torch.Tensor): 
            The input matrix.
        B (torch.Tensor): 
            The target matrix.
        check_finite (bool): 
            Whether to check that the input matrices contain only finite
            numbers. Disabling may give a performance gain, but may result in
            problems (crashes, non-termination) if the inputs do contain infinities
            or NaNs. (Default is ``True``)

    Returns:
        (Tuple[torch.Tensor, torch.Tensor]): 
            (R, scale):
                R (torch.Tensor):
                    The matrix solution of the orthogonal Procrustes problem.
                    Minimizes the Frobenius norm of ``(A @ R) - B``, subject to
                    ``R.T @ R = I``.
                scale (torch.Tensor):
                    Sum of the singular values of ``A.T @ B``.

    References:
        [1] Peter H. Schonemann, "A generalized solution of the orthogonal
        Procrustes problem", Psychometrica -- Vol. 31, No. 1, March, 1966.
        :doi:`10.1007/BF02289451`
    """
    if check_finite:
        if not torch.isfinite(A).all() or not torch.isfinite(B).all():
            raise ValueError("Input contains non-finite values.")
    assert A.shape == B.shape, 'Input matrices must have the same shape.'
    assert A.ndim == 2, 'Input matrices must be 2D.'

    U, S, V = torch.linalg.svd((B.T @ A).T, full_matrices=False)
    R = U @ V
    scale = S.sum()
    return R, scale


def dtype_to_complex(dtype: torch.dtype) -> torch.dtype:
    """
    Converts a real torch dtype to a complex dtype. \n
    Pytorch 2.2 has this, but it crashes after calling it ~100 times. \n
    RH 2024

    Args:
        dtype (torch.dtype): 
            Real dtype to convert to complex dtype.

    Returns:
        (torch.dtype): 
            complex_dtype (torch.dtype):
                Complex dtype.
    """
    map = {
        torch.float16: torch.complex32,
        torch.bfloat16: torch.complex64,
        torch.float32: torch.complex64,
        torch.float64: torch.complex128,
    }
    if dtype not in map:
        raise ValueError(f'{dtype} does not have a complex equivalent in map.')
    return map[dtype]
def dtype_to_real(dtype: torch.dtype) -> torch.dtype:
    """
    Converts a complex torch dtype to a real dtype. \n
    Pytorch 2.2 has this, but it crashes after calling it ~100 times. \n
    RH 2024

    Args:
        dtype (torch.dtype): 
            Complex dtype to convert to real dtype.

    Returns:
        (torch.dtype): 
            real_dtype (torch.dtype):
                Real dtype.
    """
    map = {
        torch.complex32: torch.float16,
        torch.complex64: torch.float32,
        torch.complex128: torch.float64,
    }
    if dtype not in map:
        raise ValueError(f'{dtype} does not have a real equivalent in map.')
    return map[dtype]


def geomspace(
    start: Union[int, float], 
    stop: Union[int, float], 
    num: int, 
    endpoint: bool = True,
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    """
    Returns numbers spaced evenly on a log scale (a geometric progression).
    Matches numpy's geomspace.

    Args:
        start (Union[int, float]): 
            The starting value of the sequence.
        stop (Union[int, float]): 
            The final value of the sequence, unless `endpoint` is False.
        num (int): 
            Number of samples to generate.
        endpoint (bool): 
            If ``True``, `stop` is the last sample. Otherwise, it is not
            included. 
        dtype (torch.dtype):
            The data type to use for the tensor. 

    Returns:
        (torch.Tensor): 
            samples (torch.Tensor): 
                The samples on a log scale.
    """
    if start <= 0:
        raise ValueError('start must be greater than 0.')
    if stop <= 0:
        raise ValueError('stop must be greater than 0.')
    if endpoint:
        return torch.logspace(
            math.log10(start),
            math.log10(stop),
            int(num),
            dtype=dtype,
            base=10.0,
        )
    else:
        gain = 10 ** (math.log10(stop / start) / (num))
        return torch.logspace(
            math.log10(start),
            math.log10(start) + math.log10(gain) * (num - 1),
            int(num),
            dtype=dtype,
            base=10.0,
        )
    

class Interp1d(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, y, xnew, out=None):
        """
        Copy pasted from: https://github.com/aliutkus/torchinterp1d/blob/master/torchinterp1d/interp1d.py
        
        Linear 1D interpolation on the GPU for Pytorch.
        This function returns interpolated values of a set of 1-D functions at
        the desired query points `xnew`.
        This function is working similarly to Matlab™ or scipy functions with
        the `linear` interpolation mode on, except that it parallelises over
        any number of desired interpolation problems.
        The code will run on GPU if all the tensors provided are on a cuda
        device.

        Parameters
        ----------
        x : (N, ) or (D, N) Pytorch Tensor
            A 1-D or 2-D tensor of real values.
        y : (N,) or (D, N) Pytorch Tensor
            A 1-D or 2-D tensor of real values. The length of `y` along its
            last dimension must be the same as that of `x`
        xnew : (P,) or (D, P) Pytorch Tensor
            A 1-D or 2-D tensor of real values. `xnew` can only be 1-D if
            _both_ `x` and `y` are 1-D. Otherwise, its length along the first
            dimension must be the same as that of whichever `x` and `y` is 2-D.
        out : Pytorch Tensor, same shape as `xnew`
            Tensor for the output. If None: allocated automatically.

        """
        # making the vectors at least 2D
        is_flat = {}
        require_grad = {}
        v = {}
        device = []
        eps = torch.finfo(y.dtype).eps
        for name, vec in {'x': x, 'y': y, 'xnew': xnew}.items():
            assert len(vec.shape) <= 2, 'interp1d: all inputs must be '\
                                        'at most 2-D.'
            if len(vec.shape) == 1:
                v[name] = vec[None, :]
            else:
                v[name] = vec
            is_flat[name] = v[name].shape[0] == 1
            require_grad[name] = vec.requires_grad
            device = list(set(device + [str(vec.device)]))
        assert len(device) == 1, 'All parameters must be on the same device.'
        device = device[0]

        # Checking for the dimensions
        assert (v['x'].shape[1] == v['y'].shape[1]
                and (
                     v['x'].shape[0] == v['y'].shape[0]
                     or v['x'].shape[0] == 1
                     or v['y'].shape[0] == 1
                    )
                ), ("x and y must have the same number of columns, and either "
                    "the same number of row or one of them having only one "
                    "row.")

        reshaped_xnew = False
        if ((v['x'].shape[0] == 1) and (v['y'].shape[0] == 1)
           and (v['xnew'].shape[0] > 1)):
            # if there is only one row for both x and y, there is no need to
            # loop over the rows of xnew because they will all have to face the
            # same interpolation problem. We should just stack them together to
            # call interp1d and put them back in place afterwards.
            original_xnew_shape = v['xnew'].shape
            v['xnew'] = v['xnew'].contiguous().view(1, -1)
            reshaped_xnew = True

        # identify the dimensions of output and check if the one provided is ok
        D = max(v['x'].shape[0], v['xnew'].shape[0])
        shape_ynew = (D, v['xnew'].shape[-1])
        if out is not None:
            if out.numel() != shape_ynew[0]*shape_ynew[1]:
                # The output provided is of incorrect shape.
                # Going for a new one
                out = None
            else:
                ynew = out.reshape(shape_ynew)
        if out is None:
            ynew = torch.zeros(*shape_ynew, device=device)

        # moving everything to the desired device in case it was not there
        # already (not handling the case things do not fit entirely, user will
        # do it if required.)
        for name in v:
            v[name] = v[name].to(device)

        # calling searchsorted on the x values.
        ind = ynew.long()

        # expanding xnew to match the number of rows of x in case only one xnew is
        # provided
        if v['xnew'].shape[0] == 1:
            v['xnew'] = v['xnew'].expand(v['x'].shape[0], -1)

        # the squeeze is because torch.searchsorted does accept either a nd with
        # matching shapes for x and xnew or a 1d vector for x. Here we would
        # have (1,len) for x sometimes 
        torch.searchsorted(v['x'].contiguous().squeeze(),
                           v['xnew'].contiguous(), out=ind)

        # the `-1` is because searchsorted looks for the index where the values
        # must be inserted to preserve order. And we want the index of the
        # preceeding value.
        ind -= 1
        # we clamp the index, because the number of intervals is x.shape-1,
        # and the left neighbour should hence be at most number of intervals
        # -1, i.e. number of columns in x -2
        ind = torch.clamp(ind, 0, v['x'].shape[1] - 1 - 1)

        # helper function to select stuff according to the found indices.
        def sel(name):
            if is_flat[name]:
                return v[name].contiguous().view(-1)[ind]
            return torch.gather(v[name], 1, ind)

        # activating gradient storing for everything now
        enable_grad = False
        saved_inputs = []
        for name in ['x', 'y', 'xnew']:
            if require_grad[name]:
                enable_grad = True
                saved_inputs += [v[name]]
            else:
                saved_inputs += [None, ]
        # assuming x are sorted in the dimension 1, computing the slopes for
        # the segments
        is_flat['slopes'] = is_flat['x']
        # now we have found the indices of the neighbors, we start building the
        # output. Hence, we start also activating gradient tracking
        with torch.enable_grad() if enable_grad else contextlib.suppress():
            v['slopes'] = (
                    (v['y'][:, 1:]-v['y'][:, :-1])
                    /
                    (eps + (v['x'][:, 1:]-v['x'][:, :-1]))
                )

            # now build the linear interpolation
            ynew = sel('y') + sel('slopes')*(
                                    v['xnew'] - sel('x'))

            if reshaped_xnew:
                ynew = ynew.view(original_xnew_shape)

        ctx.save_for_backward(ynew, *saved_inputs)
        return ynew

    @staticmethod
    def backward(ctx, grad_out):
        inputs = ctx.saved_tensors[1:]
        gradients = torch.autograd.grad(
                        ctx.saved_tensors[0],
                        [i for i in inputs if i is not None],
                        grad_out, retain_graph=True)
        result = [None, ] * 5
        pos = 0
        for index in range(len(inputs)):
            if inputs[index] is not None:
                result[index] = gradients[pos]
                pos += 1
        return (*result,)
interp1d = Interp1d.apply
    

################ Circular Statistics ################


def _circfuncs_common(samples: torch.Tensor, high: float, low: float):
    """
    Helper function for circular statistics. \n 
    This function is used to ensure that the output of a circular operation is
    always within the range [low, high). \n 
    RH 2024

    Args:
        samples (np.ndarray or torch.Tensor):
            Input values
        high (float or np.ndarray or torch.Tensor):
            High value
        low (float or np.ndarray or torch.Tensor):
            Low value

    Returns:
        output (np.ndarray or torch.Tensor):
            Output values
    """
    nan, pi = (torch.tensor(v, dtype=samples.dtype, device=samples.device) for v in [torch.nan, torch.pi])

    ## If number of elements is 0, return nan
    if samples.numel() == 0:
        return nan, nan
    
    ## sin and cos
    samples = (samples - low) * 2.0 * pi / (high - low)
    sin_samp = torch.sin(samples)
    cos_samp = torch.cos(samples)

    return sin_samp, cos_samp


def circmean(samples: torch.Tensor, high: float = 2*np.pi, low: float = 0.0, axis: Optional[List[int]] = None, nan_policy: str = 'propagate'):
    """
    Circular mean of samples. Equivalent results to scipy.stats.circmean. \n
    RH 2024

    Args:
        samples (np.ndarray or torch.Tensor):
            Input values
        high (float or np.ndarray or torch.Tensor):
            High value
        low (float or np.ndarray or torch.Tensor):
            Low value
        axis (int):
            Axis along which to take the mean
        nan_policy (str):
            Policy for handling NaN values: \n
                * 'propagate' - Propagate NaN values.
                * 'omit' - Ignore NaN values.
                * 'raise' - Raise an error if NaN values are present.

    Returns:
        mean (np.ndarray or torch.Tensor):
            Mean values
    """

    if nan_policy == 'raise':
        if torch.any(torch.isnan(samples)):
            raise ValueError("Input contains NaN values")
    
    pi = torch.tensor(torch.pi, dtype=samples.dtype, device=samples.device)
    sin_samp, cos_samp = _circfuncs_common(samples, high, low)

    if nan_policy == 'omit':
        sin_sum = torch.nansum(sin_samp, dim=axis)
        cos_sum = torch.nansum(cos_samp, dim=axis)
    elif nan_policy == 'propagate':
        sin_sum = torch.sum(sin_samp, dim=axis)
        cos_sum = torch.sum(cos_samp, dim=axis)
    elif nan_policy == 'raise':
        sin_sum = torch.sum(sin_samp, dim=axis)
        cos_sum = torch.sum(cos_samp, dim=axis)
    else:
        raise ValueError("Invalid nan_policy")
    
    res = torch.arctan2(sin_sum, cos_sum)

    if res.ndim == 0:
        res = res + 2 * pi if res < 0 else res
    else:
        res[res < 0] += 2 * pi
    
    # res = res[()]

    return res*(high - low)/2.0/pi + low


def cirvar(samples: torch.Tensor, high=2*np.pi, low=0, axis=None, nan_policy='propagate'):
    """
    Circular variance of samples. Equivalent results to scipy.stats.circvar. \n
    RH 2024

    Args:
        samples (np.ndarray or torch.Tensor):
            Input values
        high (float or np.ndarray or torch.Tensor):
            High value
        low (float or np.ndarray or torch.Tensor):
            Low value
        axis (int):
            Axis along which to take the variance
        nan_policy (str):
            Policy for handling NaN values. Can only be 'propagate' for now.

    Returns:
        variance (np.ndarray or torch.Tensor):
            Variance values
    """

    if nan_policy != 'propagate':
        raise NotImplementedError("Only 'propagate' nan_policy is supported")
        
    sin_samp, cos_samp = _circfuncs_common(samples, high, low)
    sin_mean = sin_samp.mean(axis)
    cos_mean = cos_samp.mean(axis)
    
    R = torch.sqrt(sin_mean**2 + cos_mean**2)

    return 1 - R


def circstd(samples: torch.Tensor, high: float = 2*np.pi, low: float = 0.0, axis: Optional[List[int]] = None, nan_policy: str = 'propagate', normalize: bool = False):
    """
    Circular standard deviation of samples. Equivalent results to
    scipy.stats.circstd. \n 
    RH 2024

    Args:
        samples (np.ndarray or torch.Tensor):
            Input values
        high (float or np.ndarray or torch.Tensor):
            High value
        low (float or np.ndarray or torch.Tensor):
            Low value
        axis (int):
            Axis along which to take the standard deviation
        nan_policy (str):
            Policy for handling NaN values. Can only be 'propagate' for now.
        normalize (bool):
            Whether to normalize the standard deviation. If True, the result is
            equal to ``sqrt(-2*log(R))`` and does not depend on the variable
            units. If False (default), the returned value is scaled by
            ``((high-low)/(2*pi))``.


    Returns:
        std (np.ndarray or torch.Tensor):
            Standard deviation values
    """

    if nan_policy != 'propagate':
        raise NotImplementedError("Only 'propagate' nan_policy is supported")
    
    pi = torch.tensor(np.pi, dtype=samples.dtype, device=samples.device)

    sin_samp, cos_samp = _circfuncs_common(samples, high, low)
    sin_mean = sin_samp.mean(axis)
    cos_mean = cos_samp.mean(axis)
    R = torch.sqrt(sin_mean**2 + cos_mean**2)

    res = torch.sqrt(-2*torch.log(R))
    if not normalize:
        res *= (high-low)/(2.*pi)
    return res


##################################################################
######################### TENSORLY ###############################
##################################################################

def tensorly_cp_to_device(cp, device='cpu'):
    """
    Moves the factors and weights of a tensorly cp object to a particular
    device.

    RH 2024

    Args:
        cp (tensorly.cp_tensor.CP):
            The tensorly CP object to move to a device.
        device (str):
            The device to move the factors and weights to.

    Returns:
        cp (tensorly.cp_tensor.CP):
            The tensorly CP object with factors and weights moved to the device.
    """
    for ii in range(len(cp.factors)):
        cp.factors[ii] = cp.factors[ii].to(device)
    cp.weights = cp.weights.to(device)
    return cp


from typing import Dict, Type, Any, Union, Optional, Callable, Tuple, List
import time
import logging
import os

import numpy as np
import torch
import optuna
from tqdm.auto import tqdm

class Convergence_checker:
    """
    Checks for convergence during an optimization process. Uses Ordinary Least
    Squares (OLS) to fit a line to the last 'window_convergence' number of
    iterations.\n
    RH 2022
    
    Args:
        tol_convergence (float): 
            Tolerance for convergence.\n
            Corresponds to the slope of the line that is fit.\n
            If fractional==True, then tol_convergence is the fractional
            change in loss over the window_convergence
        fractional (bool):
            If True, then tol_convergence is the fractional change in loss
            over the window_convergence. ie: delta(lossWin) / mean(lossWin)
        window_convergence (int):
            Number of iterations to use for fitting the line.
        mode (str):
            Where deltaLoss = loss[current] - loss[window convergence steps ago]\n
            For typical loss curves, deltaLoss should be negative. So common
            modes are: 'greater' with tol_convergence = -1e-x, and 'less'
            with tol_convergence = 1e-x.
            Mode for how criterion is defined:\n
                * 'less': converged = deltaLoss < tol_convergence (default)
                * 'abs_less': converged = abs(deltaLoss) < tol_convergence
                * 'greater': converged = deltaLoss > tol_convergence
                * 'abs_greater': converged = abs(deltaLoss) > tol_convergence
                * 'between': converged = tol_convergence[0] < deltaLoss < tol_convergence[1]\n
                    (tol_convergence must be a list or tuple, if mode='between')
        max_iter (int):
            Maximum number of iterations to run for.\n
            If None, then no maximum.
        max_time (float):
            Maximum time to run for (in seconds).\n
            If None, then no maximum.
        nan_policy (str):
            Policy for handling NaNs in the loss history.\n
                * 'omit': (default) Ignore NaNs in the loss history.
                * 'halt': Return converged=True if NaNs are found in the loss history.
                * 'allow': Allow NaNs in the loss history.
                * 'raise': Raise an error if NaNs are found in the loss history.
        explode_tolerance (float):
            Tolerance for exploding loss. If the ratio in loss between the
            current and previous iteration (current / previous) is greater than
            this value, then the optimization is considered to have exploded. If
            None, then no explosion check is done.
        explode_patience (int):
            Number of iterations to wait before checking for explosion. If None, then
            no patience is used.

    """
    def __init__(
        self,
        tol_convergence=-1e-2,
        fractional=False,
        window_convergence=100,
        mode='greater',
        max_iter=None,
        max_time=None,
        nan_policy='omit',
        explode_tolerance=None,
        explode_patience=10,
    ):
        """
        Initialize the convergence checker.
        """
        self.window_convergence = window_convergence
        self.tol_convergence = tol_convergence
        self.fractional = fractional

        self.line_regressor = torch.cat((torch.linspace(0,1,window_convergence)[:,None], torch.ones((window_convergence,1))), dim=1)

        if mode=='less':          self.fn_criterion = (lambda diff: diff < self.tol_convergence)
        elif mode=='abs_less':    self.fn_criterion = (lambda diff: abs(diff) < self.tol_convergence)
        elif mode=='greater':     self.fn_criterion = (lambda diff: diff > self.tol_convergence)
        elif mode=='abs_greater': self.fn_criterion = (lambda diff: abs(diff) > self.tol_convergence)
        elif mode=='between':     self.fn_criterion = (lambda diff: self.tol_convergence[0] < diff < self.tol_convergence[1])
        assert self.fn_criterion is not None, f"mode '{mode}' not recognized"

        self.max_iter = max_iter
        self.max_time = max_time
        self.nan_policy = nan_policy
        self.explode_tolerance = explode_tolerance
        self.explode_patience = explode_patience

        self.iter = -1

    def OLS(self, y, X=None):
        """
        Ordinary least squares.
        Fits a line and bias term (stored in self.line_regressor) to y input.
        """
        X = self.line_regressor if X is None else X
        theta = torch.inverse(X.T @ X) @ X.T @ y
        y_rec = X @ theta
        bias = theta[-1]
        theta = theta[:-1]

        return theta, y_rec, bias

    def __call__(
        self,
        loss_history=None,
        loss_single=None,
    ):
        """
        Forward pass of the convergence checker.
        Checks if the last 'window_convergence' number of iterations are
         within 'tol_convergence' of the line fit.

        Args:
            loss_history (list or array):
                List of loss values for entire optimization process.
                If None, then internally tracked loss_history is used.
            loss_single (float):
                Single loss value for current iteration.

        Returns:
            delta_window_convergence (float):
                Difference between first and last element of the fit line
                 over the range of 'window_convergence'.
                 diff_window_convergence = (y_rec[-1] - y_rec[0])
            loss_smooth (float):
                The mean loss over 'window_convergence'.
            converged (bool):
                True if the 'diff_window_convergence' is less than
                 'tol_convergence'.
        """
        ## Initialize timer t0
        if self.iter == 0:
            self.t0 = time.time()
        
        ## Increment iter
        self.iter += 1

        ## Parse args and prepare loss_history
        if loss_history is None:
            if not hasattr(self, 'loss_history'):
                assert loss_single is not None, "loss_history and loss_single are both None"
                self.loss_history = []
            self.loss_history.append(loss_single)
            loss_history = self.loss_history

        ## Check for explosion
        if (self.explode_tolerance is not None) and (self.iter > self.explode_patience):
            if abs(loss_history[-1] / loss_history[-2]) > self.explode_tolerance:
                return torch.nan, torch.nan, True

        ## Prepare loss_window
        loss_window = torch.as_tensor(loss_history[-self.window_convergence:], device='cpu', dtype=torch.float32)

        ## Handle nan_policy, fit line, and make loss_smooth
        if self.nan_policy=='omit':
            loss_window = loss_window[~torch.isnan(loss_window)]
            loss_smooth = torch.nanmean(loss_window)

            ## Wait until window_convergence number of iterations
            if len(loss_history) < self.window_convergence:
                return torch.nan, loss_smooth, False

            theta, y_rec, bias = self.OLS(y=loss_window, X=self.line_regressor[~torch.isnan(loss_window)])
        else:
            if self.nan_policy=='halt':
                if torch.isnan(loss_window).any():
                    return torch.nan, torch.nan, True
            elif self.nan_policy=='raise':
                if torch.isnan(loss_window).any():
                    raise ValueError("NaNs found in loss history")
            elif self.nan_policy=='allow':
                pass

            ## Wait until window_convergence number of iterations
            if len(loss_history) < self.window_convergence:
                return torch.nan, torch.nan, False
            
            loss_smooth = torch.mean(loss_window)
            theta, y_rec, bias = self.OLS(y=loss_window)

        ## Check for convergence
        delta_window_convergence = (y_rec[-1] - y_rec[0]) if not self.fractional else (y_rec[-1] - y_rec[0]) / ((y_rec[-1] + y_rec[0])/2)
        converged = self.fn_criterion(delta_window_convergence)

        ## Check for max_iter and max_time
        if self.max_iter is not None:
            converged = converged or (len(loss_history) >= self.max_iter)
        if self.max_time is not None:
            converged = converged or (time.time() - self.t0 > self.max_time)
        
        return delta_window_convergence.item(), loss_smooth.item(), converged
    

class Convergence_checker_optuna:
    """
    Checks if the optuna optimization has converged.
    RH 2023

    Args:
        n_patience (int): 
            Number of trials to look back to check for convergence. 
            Also the minimum number of trials that must be completed 
            before starting to check for convergence. 
            (Default is *10*)
        tol_frac (float): 
            Fractional tolerance for convergence. 
            The best output value must change by less than this 
            fractional amount to be considered converged. 
            (Default is *0.05*)
        max_trials (int): 
            Maximum number of trials to run before stopping. 
            (Default is *350*)
        max_duration (float): 
            Maximum number of seconds to run before stopping. 
            (Default is *600*)
        value_stop (Optional[float]):
            Value at which to stop the optimization. If the best value is equal
            to or less than this value, the optimization will stop.
            (Default is *None*)
        verbose (bool): 
            If ``True``, print messages. 
            (Default is ``True``)

    Attributes:
        bests (List[float]):
            List to hold the best values obtained in the trials.
        best (float):
            Best value obtained among the trials. Initialized with infinity.

    Example:
        .. highlight:: python
        .. code-block:: python

            # Create a ConvergenceChecker instance
            convergence_checker = ConvergenceChecker(
                n_patience=15, 
                tol_frac=0.01, 
                max_trials=500, 
                max_duration=60*20, 
                verbose=True
            )
            
            # Assume we have a study and trial objects from optuna
            # Use the check method in the callback
            study.optimize(objective, n_trials=100, callbacks=[convergence_checker.check])    
    """
    def __init__(
        self, 
        n_patience: int = 10, 
        tol_frac: float = 0.05, 
        max_trials: int = 350, 
        max_duration: float = 60*10, 
        value_stop: Optional[float] = None,
        verbose: bool = True,
    ):
        """
        Initializes the ConvergenceChecker with the given parameters.
        """
        self.bests = []
        self.best = np.inf
        self.n_patience = n_patience
        self.tol_frac = tol_frac
        self.max_trials = max_trials
        self.max_duration = max_duration
        self.value_stop = value_stop
        self.num_trial = 0
        self.verbose = verbose

        self.reason_converged = 'not converged'
        
    def check(
        self, 
        study: object, 
        trial: object,
    ):
        """
        Checks if the optuna optimization has converged. This function should be
        used as the callback function for the optuna study.

        Args:
            study (optuna.study.Study): 
                Optuna study object.
            trial (optuna.trial.FrozenTrial): 
                Optuna trial object.
        """
        dur_first, dur_last = study.trials[0].datetime_complete, trial.datetime_complete
        if (dur_first is not None) and (dur_last is not None):
            duration = (dur_last - dur_first).total_seconds()
        else:
            duration = 0
        
        if trial.value < self.best:
            self.best = trial.value
        self.bests.append(self.best)
            
        bests_recent = np.unique(self.bests[-self.n_patience:])
        if self.best == 0:
            print(f'Stopping. Best value is 0.') if self.verbose else None
            self.reason_converged = 'best value is 0'
            study.stop()
        elif self.num_trial > self.n_patience and ((np.abs(bests_recent.max() - bests_recent.min()) / np.abs(self.best)) < self.tol_frac):
            print(f'Stopping. Convergence reached. Best value ({self.best*10000}) over last ({self.n_patience}) trials fractionally changed less than ({self.tol_frac})') if self.verbose else None
            self.reason_converged = 'fractional change in best value'
            study.stop()
        elif self.num_trial >= self.max_trials:
            print(f'Stopping. Trial number limit reached. num_trial={self.num_trial}, max_trials={self.max_trials}.') if self.verbose else None
            self.reason_converged = 'max trials reached'
            study.stop()
        elif duration > self.max_duration:
            print(f'Stopping. Duration limit reached. study.duration={duration}, max_duration={self.max_duration}.') if self.verbose else None
            self.reason_converged = 'max duration reached'
            study.stop()

        if self.value_stop is not None:
            if self.best <= self.value_stop:
                print(f'Stopping. Best value ({self.best}) is less than or equal to value_stop ({self.value_stop}).') if self.verbose else None
                self.reason_converged = 'value stop'
                study.stop()
            
        if self.verbose:
            print(f'Trial num: {self.num_trial}. Duration: {duration:.3f}s. Best value: {self.best:3e}. Current value:{trial.value:3e}') if self.verbose else None
        self.num_trial += 1


class OptunaProgressBar:
    """
    A customizable progress bar for Optuna's study.optimize().

    Args:
        n_trials (int, optional): 
            The number of trials. Required if timeout is not set.
        timeout (float, optional): 
            The maximum time to run in seconds. Required if n_trials is not set.
        tqdm_kwargs (dict, optional): 
            Additional keyword arguments to pass to tqdm.
    """

    def __init__(self, n_trials=None, timeout=None, **tqdm_kwargs):
        from optuna import logging as optuna_logging

        self._n_trials = n_trials
        self._timeout = timeout
        self._tqdm_kwargs = tqdm_kwargs

        # Read TQDM environment variables
        self._load_env_variables()

        # Initialize progress bar
        self._progress_bar = None
        self._last_elapsed_seconds = 0.0

        # Setup logging handler to redirect Optuna logs to tqdm
        self._tqdm_handler = _TqdmLoggingHandler()
        self._tqdm_handler.setLevel(logging.INFO)
        self._tqdm_handler.setFormatter(optuna_logging.create_default_formatter())
        optuna_logging.disable_default_handler()
        optuna_logging._get_library_root_logger().addHandler(self._tqdm_handler)

    def __call__(self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
        # Initialize progress bar on first call
        if self._progress_bar is None:
            if self._n_trials is not None:
                self._progress_bar = tqdm(total=self._n_trials, **self._tqdm_kwargs)
            elif self._timeout is not None:
                total = tqdm.format_interval(self._timeout)
                fmt = "{desc} {percentage:3.0f}%|{bar}| {elapsed}/" + total
                self._progress_bar = tqdm(total=self._timeout, bar_format=fmt, **self._tqdm_kwargs)
            else:
                raise ValueError("Either n_trials or timeout must be set.")

        # Update progress bar
        if self._n_trials is not None:
            self._progress_bar.update(1)
        elif self._timeout is not None:
            elapsed = study._stop_watch.elapsed_time()
            time_diff = elapsed - self._last_elapsed_seconds
            self._progress_bar.update(time_diff)
            self._last_elapsed_seconds = elapsed

        # Update description with best trial information
        if not study._is_multi_objective():
            try:
                best_value = study.best_value
                self._progress_bar.set_description(f"Best value: {best_value:.6g}")
            except ValueError:
                pass  # No trials completed yet

    def close(self):
        from optuna import logging as optuna_logging
        if self._progress_bar is not None:
            self._progress_bar.close()
        # Restore Optuna's default logging handler
        optuna_logging._get_library_root_logger().removeHandler(self._tqdm_handler)
        optuna_logging.enable_default_handler()

    def _load_env_variables(self):
        # Load TQDM environment variables
        env_vars = {
            'disable': os.environ.get('TQDM_DISABLE', '').lower() == 'true',
            'mininterval': float(os.environ.get('TQDM_MININTERVAL', 0.1)),
            'maxinterval': float(os.environ.get('TQDM_MAXINTERVAL', 10.0)),
            'miniters': int(os.environ.get('TQDM_MINITERS', 1)),
            'smoothing': float(os.environ.get('TQDM_SMOOTHING', 0.3)),
        }
        # Update tqdm_kwargs with environment variables if not already set
        for key, value in env_vars.items():
            if key not in self._tqdm_kwargs:
                self._tqdm_kwargs[key] = value
class _TqdmLoggingHandler(logging.StreamHandler):
    """Logging handler to redirect Optuna logs to tqdm.write()."""
    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


import re
import copy
from collections.abc import MutableMapping
from typing import Dict, Any, Optional, Union, List, Tuple, Callable, Iterable, Iterator, Type
import itertools

import toolz

"""
This module is intended to have minimal dependencies.
It is called by server.py which is intended to run
 outside of a specialized environment, so no weird
 libraries should be required.
RH 2022
"""

class Lazy_repeat_obj(itertools.repeat):
    """
    Makes a lazy iterator that repeats an object.
    Makes a subclass of itertools.repeat that has a length.
    RH 2021

    Args:
        obj (Any):
            Object to repeat.
        pseudo_length (int):
            length of the iterator. Don't use keyword argument, just pass the
            length as the second argument.
    """
    def __init__(
        self, 
        obj: Any, 
        pseudo_length: Optional[int] = None,
    ):
        """
        Initializes the lazy iterator.
        """
        self.obj = obj
        self.pseudo_length = pseudo_length
        self.count = 0

    def __iter__(self):
        """
        Returns the iterator.
        """
        return self
    
    def __next__(self):
        """
        Returns the next object.
        """
        if self.pseudo_length is not None:
            if self.count >= self.pseudo_length:
                raise StopIteration
            self.count += 1
        return self.obj
    
    def __len__(self):
        """
        Returns the length of the iterator.
        """
        return self.pseudo_length
    
    def __repr__(self):
        """
        Returns the representation of the object.
        """
        return f"lazy_repeat_obj({self.obj}, pseudo_length={self.pseudo_length})"


def flatten_list(irregular_list):
    """
    Flattens a list of lists into a single list.
    Stolen from https://stackabuse.com/python-how-to-flatten-list-of-lists/
    RH 2022

    Args:
        irregular_list (list):
            list of lists to flatten

    Returns:
        output (list):
            flattened list
    """
    helper = lambda irregular_list:[element for item in irregular_list for element in flatten_list(item)] if type(irregular_list) is list else [irregular_list]
    return helper(irregular_list)


def flatten_dict(d: MutableMapping, parent_key: str = '', sep: str ='.') -> MutableMapping:
    """
    Flattens a dictionary of dictionaries into a 
     single dictionary.
    NOTE: Turns all keys into strings.
    Stolen from https://stackoverflow.com/a/6027615
    RH 2022

    Args:
        d (Dict):
            dictionary to flatten
        parent_key (str):
            key to prepend to flattened keys
            USED INTERNALLY FOR RECURSION
        sep (str):
            separator to use between keys
            USED INTERNALLY FOR RECURSION

    Returns:
        output (Dict):
            flattened dictionary
    """

    items = []
    for k, v in d.items():
        new_key = str(parent_key) + str(sep) + str(k) if parent_key else str(k)
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def deep_update_dict(dictionary, key, new_val=None, new_key=None, in_place=False):
    """
    Updates a dictionary with a new value.
    RH 2022

    Args:
        dictionary (Dict):
            dictionary to update
        key (list of str):
            Key to update
            List elements should be strings.
            Each element should be a hierarchical
             level of the dictionary.
            DEMO:
                deep_update_dict(params, ['dataloader_kwargs', 'prefetch_factor'], val)
        new_val (any):
            If not None, the value to update with this
            If None, then new_key must be specified and will only
             be used to update the key.
        new_key (str):
            If not None, the key will be updated with this key.
             [key[-1]] will be deleted and replaced with new_key.
            If None, then [key[-1]] will be updated with new_val.
             
        in_place (bool):
            whether to update in place

    Returns:
        output (Dict):
            updated dictionary
    """
    def helper_deep_update_dict(d, key):
        if type(key) is str:
            key = [key]

        assert key[0] in d, f"RH ERROR, key: '{key[0]}' is not found"

        if type(key) is list:
            if len(key) > 1:
                helper_deep_update_dict(d[key[0]], key[1:])
            elif len(key) == 1:
                val = d[key[0]] if new_val is None else new_val
                if new_key is None:
                    d[key[0]] = val
                else:
                    d[new_key] = val
                    del d[key[0]]

    if in_place:
        helper_deep_update_dict(dictionary, key)
    else:
        d = copy.deepcopy(dictionary)
        helper_deep_update_dict(d, key)
        return d


def dict_shared_items(
    d1, 
    d2, 
    fn_check_equality=None, 
    error_if_equality_check_fails=True, 
    verbose=False
):
    """
    Returns the matching items between two dictionaries.
    Searches for d1 items within d2.
    RH 2023

    Args:
        d1 (Dict):
            first dictionary
        d2 (Dict):
            second dictionary
        fn_check_equality (function):
            function to use to check equality.
            If None, then equality is checked with ==.
            If not None, then the function should take two
             arguments and return True if they are equal.
        error_if_equality_check_fails (bool):
            whether to raise an error if the equality check fails.
            If False and verbose==True, then a warning is printed
             instead.
        verbose (bool):
            whether to print warnings if the equality check fails.

    Returns:
        output (Dict):
            dictionary of matching items
    """
    def check_equality(v1, v2):
        try:
            if fn_check_equality is not None:
                out = fn_check_equality(v1, v2)
            else:
                if v1 == v2:
                    out = True
                else:
                    out = False
        except ValueError as e:
            if error_if_equality_check_fails:
                raise e
            else:
                print(f'WARNING: {e}') if verbose else None
                out = False
        return out
    
    return {k: d1[k] for k in d1 if k in d2 and check_equality(d1[k], d2[k])}
def dict_diff_items(d1, d2):
    """
    Returns the differing items between two dictionaries.
    Searches for d1 items within d2.
    RH 2022
    """
    return {k: d1[k] for k in d1 if k in d2 and d1[k] != d2[k]}
def dict_missing_keys(d1, d2):
    """
    Returns the keys in d1 that are missing in d2.
    Symmetric difference between d1 and d2 keys.
    RH 2022
    """
    return set(d1.keys()).symmetric_difference(set(d2.keys()))


def find_differences_across_dictionaries(dicts):
    """
    Finds differences across many dictionaries.
    RH 2022

    Args:
        dicts (List):
            List of dictionaries to compare.

    Returns:
        params_unchanging (list of dicts):
            List of dictionary items that are the 
             same across all dictionaries.
        params_changing (list of dicts):
            List of dictionary items that are 
             different in at least one dictionary.
    """
    import numpy as np

    def get_binary_search_combos(n):
        combos = list(np.arange(n))
        if len(combos)%2 == 1:
            combos.append(combos[-1])
        combos = np.array(combos).reshape(len(combos)//2, 2)
        return combos

    ## flatten params to ease matching functions
    params_flat = [flatten_dict(param) for param in dicts]

    ## assert that set of all keys in each dict is exactly the same
    sets_of_keys = set([tuple(sorted(param.keys())) for param in params_flat])
    if len(sets_of_keys) > 1:
        import itertools
        ## use itertools to compare all combinations of sets_of_keys and find the differences
        combos = list(itertools.combinations(sets_of_keys, 2))
        diffs = []
        for combo in combos:
             diffs.append(set(combo[0]).symmetric_difference(set(combo[1])))
        ## collapse diffs into a single set
        diffs = set([item for sublist in diffs for item in sublist])
    assert len(sets_of_keys) == 1, f"RH ERROR, the following keys are not the same across all dictionaries: {diffs}"

    ## find unchanging params
    params_unchanging = copy.deepcopy(params_flat)
    while len(params_unchanging) > 1:
        combos = get_binary_search_combos(len(params_unchanging))
        params_unchanging = [dict_shared_items(params_unchanging[combo[0]], params_unchanging[combo[1]]) for combo in combos]
    params_unchanging = params_unchanging[0]

    ## find keys that are not unchanging
    mk = dict_missing_keys(params_flat[0], params_unchanging)

    ## make list dicts of changing params
    params_changing = [{k: params[k] for k in mk} for params in params_flat]
    
    return params_unchanging, params_changing


def find_subDict_key(d: dict, s: str, max_depth: int=9999999):
    """
    Recursively search for a sub-dictionary that contains the given string.
    Yield the result.

    Args:
        d (dict):
            dictionary to search
        s (str):
            string of the key to search for using regex
        max_depth (int):
            maximum depth to search.
            1 means only search the keys in the top level of
             the dictionary. 2 means the first and second level.

    Returns:
        k_all (list of tuples):
            List of 2-tuples: (list of tuples containing:
             list of strings of keys to sub-dictionary, value
             of sub-dictionary)
    """
    def helper_find_subDict_key(d, s, depth=999, _k_all=[]):
        """
        _k_all: 
            Used for recursion. List of keys. Set to [] on first call.
        depth: 
            Used for recursion. Decrements by 1 each call. At 0, stops
             recursion.
        """
        if depth > 0:    
            depth -= 1
            for k, v in d.items():
                if re.search(s, k):
                    yield _k_all + [k], v
                if isinstance(v, dict):
                    yield from helper_find_subDict_key(v, s, depth, _k_all + [k])
    return list(helper_find_subDict_key(d, s, depth=max_depth, _k_all=[]))


def merge_dicts(dicts: List[Dict]) -> Dict:
    """
    Merges a list of dictionaries into a single dictionary.
    If there are sub-dictionaries, then they are merged recursively.
    RH 2023

    Args:
        dicts (List):
            List of dictionaries to merge.

    Returns:
        output (Dict):
            merged dictionary
    """
    assert isinstance(dicts, list), f"RH ERROR, dicts must be a list, not {type(dicts)}"

    merged_dict = {}
    for d in dicts:
        for k, v in d.items():
            if isinstance(v, dict):
                merged_dict[k] = merge_dicts([merged_dict.get(k, {}), v])
            else:
                merged_dict[k] = v

    return merged_dict


def invert_dict(d: Dict) -> Dict:
    """
    Inverts a dictionary. Requires that values are hashable. 
    Warning: if values are not unique, then only the last key will be kept.
    RH 2024

    Args:
        d (Dict):
            dictionary to invert

    Returns:
        output (Dict):
            inverted dictionary
    """
    return {v: k for k, v in d.items()}


def serializable_dict(
    obj: Any,
    allowed_libraries: Optional[List[str]] = [
        'bnpm',
        'builtins',
        'collections',
        'datetime',
        'itertools',
        'math',
        'numbers',
        'os',
        'pathlib',
        'string',
        'time',
        'numpy',
        'scipy',
        'sklearn',
    ],
) -> Dict[str, Any]:
    """
    Returns a serializable dictionary that can be saved to disk. This method
    goes through all items in self.__dict__ and checks if they are
    serializable. If they are, add them to a dictionary to be returned.

    Args:
        obj (Any):
            Object to serialize.

    Returns:
        (Dict[str, Any]): 
            serializable_dict (Dict[str, Any]): 
                Dictionary containing serializable items.
    """
    from functools import partial
    ## Go through all items in self.__dict__ and check if they are serializable.
    ### If they are, add them to a dictionary to be returned.
    import pickle

    ## Define a list of libraries and classes that are allowed to be serialized.
    # allowed_libraries = 
    def is_library_allowed(obj):
        try:
            try:
                module_name = obj.__module__.split('.')[0]
            except:
                success = False
            try:
                module_name = obj.__class__.__module__.split('.')[0]
            except:
                success = False
        except:
            success = False
        else:
            ## Check if the module_name is in the allowed_libraries list.
            if module_name in allowed_libraries:
                success = True
            else:
                success = False
        return success

    def make_serializable_dict(obj, depth=0, max_depth=100, name=None):
        """
        Recursively go through all items in self.__dict__ and check if they are serializable.
        """
        msd_partial = partial(make_serializable_dict, depth=depth+1, max_depth=max_depth)
        if depth > max_depth:
            raise Exception(f'RH ERROR: max_depth of {max_depth} reached with object: {obj}')

        serializable_dict = {}
        if hasattr(obj, '__dict__') and is_library_allowed(obj):
            for key, val in obj.__dict__.items():
                try:
                    serializable_dict[key] = msd_partial(val, name=key)
                except:
                    pass

        elif isinstance(obj, (list, tuple, set, frozenset)):
            serializable_dict = [msd_partial(v, name=f'{name}_{ii}') for ii,v in enumerate(obj)]
        elif isinstance(obj, dict):
            serializable_dict = {k: msd_partial(v, name=f'{name}_{k}') for k,v in obj.items()}
        else:
            try:
                assert is_library_allowed(obj), f'RH ERROR: object {obj} is not serializable'
                pickle.dumps(obj)
            except:
                return {'__repr__': repr(obj)} if hasattr(obj, '__repr__') else {'__str__': str(obj)} if hasattr(obj, '__str__') else None

            serializable_dict = obj

        return serializable_dict

    serializable_dict = make_serializable_dict(obj, depth=0, max_depth=100, name='object')
    return serializable_dict


def recursive_valmap(
    func: Callable[[Any], Any],
    d: Dict,
    depth: int = -1,
):
    """
    Apply a function to all values in a nested dictionary. \n
    If item is a dictionary, then apply the function to all values in the
    dictionary. \n
    Uses toolz.dicttoolz.valmap. \n
    RH 2024

    Args:
        func (Callable):
            Function to apply to all values.
        d (Dict):
            Dictionary to apply the function to.
        factory (Type[Dict]):
            Type of dictionary to use for the output.
        depth (int):
            Maximum depth to apply the function.
            If -1, then apply to all depths.
            If 1, then apply to the top level only.
            If 0, then return the dictionary as is.

    Returns:
        output (Dict):
            Dictionary with the function applied to all values.
    """
    def helper_recursive_valmap(func, d, depth):
        if depth == 0:
            return d
        return toolz.valmap(
            lambda x: helper_recursive_valmap(func, x, depth-1) if isinstance(x, dict) else func(x),
            d
        )

    return helper_recursive_valmap(func, d, depth)


############################################################
################# PARAMETER DICTIONARIES ###################
############################################################


def fill_in_dict(
    d: Dict, 
    defaults: Dict,
    verbose: bool = True,
    hierarchy: List[str] = ['dict'], 
):
    """
    In-place. Fills in dictionary ``d`` with values from ``defaults`` if they
    are missing. Works hierachically.
    RH 2023

    Args:
        d (Dict):
            Dictionary to fill in.
            In-place.
        defaults (Dict):
            Dictionary of defaults.
        verbose (bool):
            Whether to print messages.
        hierarchy (List[str]):
            Used internally for recursion.
            Hierarchy of keys to d.
    """
    from copy import deepcopy
    for key in defaults:
        if key not in d:
            print(f"Key '{key}' not found in params dictionary: {' > '.join([f'{str(h)}' for h in hierarchy])}. Using default value: {defaults[key]}") if verbose else None
            d.update({key: deepcopy(defaults[key])})
        elif isinstance(defaults[key], dict):
            assert isinstance(d[key], dict), f"Key '{key}' is a dict in defaults, but not in params. {' > '.join([f'{str(h)}' for h in hierarchy])}."
            fill_in_dict(d[key], defaults[key], hierarchy=hierarchy+[key])
            

def check_keys_subset(d, default_dict, hierarchy=['defaults']):
    """
    Checks that the keys in d are all in default_dict. Raises an error if not.
    RH 2023

    Args:
        d (Dict):
            Dictionary to check.
        default_dict (Dict):
            Dictionary containing the keys to check against.
        hierarchy (List[str]):
            Used internally for recursion.
            Hierarchy of keys to d.
    """
    default_keys = list(default_dict.keys())
    for key in d.keys():
        assert key in default_keys, f"Parameter '{key}' not found in defaults dictionary: {' > '.join([f'{str(h)}' for h in hierarchy])}."
        if isinstance(default_dict[key], dict) and isinstance(d[key], dict):
            check_keys_subset(d[key], default_dict[key], hierarchy=hierarchy+[key])


def prepare_params(params, defaults, verbose=True):
    """
    Does the following:
        * Checks that all keys in ``params`` are in ``defaults``.
        * Fills in any missing keys in ``params`` with values from ``defaults``.
        * Returns a deepcopy of the filled-in ``params``.

    Args:
        params (Dict):
            Dictionary of parameters.
        defaults (Dict):
            Dictionary of defaults.
        verbose (bool):
            Whether to print messages.
    """
    from copy import deepcopy
    ## Check inputs
    assert isinstance(params, dict), f"p must be a dict. Got {type(params)} instead."
    ## Make sure all the keys in p are valid
    check_keys_subset(params, defaults)
    ## Fill in any missing keys with defaults
    params_out = deepcopy(params)
    fill_in_dict(params_out, defaults, verbose=verbose)

    return params_out


def make_grid_search_dicts(search_space):
    """
    Makes a grid sweep over the search space.
    RH 2024

    Args:
        search_space (dict):
            Dictionary with keys as parameter names and values as lists of
            parameter values.\n
            Example: {'lr': [0.1, 0.01, 0.001], 'batch_size': [32, 64]}

    Returns:
        ss_comb_dicts (list):
            List of dictionaries, where each dictionary is a combination of
            parameter values from the search space.\n
            Example: [{'lr': 0.1, 'batch_size': 32}, {'lr': 0.1, 'batch_size': 64}, ...]
    """
    import itertools
    vals_comb = list(itertools.product(*search_space.values()))
    ss_comb_dicts = [{k: val for k, val in zip(search_space.keys(), vals)} for vals in vals_comb]
    return ss_comb_dicts


from typing import Union, Tuple, List, Dict, Any, Optional
import functools
import math

import scipy.signal
import scipy.stats
import numpy as np
import matplotlib.pyplot as plt
import torch
import opt_einsum

from . import circular
from . import misc
from . import torch_helpers
from . import timeSeries
from . import indexing


def design_butter_bandpass(lowcut, highcut, fs, order=5, plot_pref=False):
    '''
    designs a butterworth bandpass filter.
    Makes a lowpass filter if lowcut is 0.
    Makes a highpass filter if highcut is fs/2.
    RH 2021

        Args:
            lowcut (scalar): 
                frequency (in Hz) of low pass band
            highcut (scalar):  
                frequency (in Hz) of high pass band
            fs (scalar): 
                sample rate (frequency in Hz)
            order (int): 
                order of the butterworth filter
        
        Returns:
            b (ndarray): 
                Numerator polynomial coeffs of the IIR filter
            a (ndarray): 
                Denominator polynomials coeffs of the IIR filter
    '''
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq

    if low <= 0:
        ## Make a lowpass filter
        b, a = scipy.signal.butter(N=order, Wn=high, btype='low')
    elif high >= 1:
        ## Make a highpass filter
        b, a = scipy.signal.butter(N=order, Wn=low, btype='high')
    else:
        b, a = scipy.signal.butter(N=order, Wn=[low, high], btype='band')
    
    if plot_pref:
        plot_digital_filter_response(b=b, a=a, fs=fs, worN=100000)
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, axis=-1, order=5, plot_pref=False):
    '''
    applies a butterworth bandpass filter
    RH 2021
    
        Args:
            data (ndarray): 
                data array. filtering done on 
                 defined axis 
            lowcut (scalar): 
                frequency (in Hz) of low pass band
            highcut (scalar): 
                frequency (in Hz) of high pass band
            fs (scalar): 
                sample rate (frequency in Hz)
            order (int): 
                order of the butterworth filter
        
        Returns:
            y (ndarray): 
                filtered data array
    '''
    b, a = design_butter_bandpass(lowcut, highcut, fs, order=order, plot_pref=plot_pref)
    y = scipy.signal.lfilter(b, a, data, axis=axis)
    return y


def design_fir_bandpass(lowcut, highcut, num_taps=30001, q=None, fs=30, window='hamming', plot_pref=True):
    '''
    designs a FIR bandpass filter.
    Makes a lowpass filter if lowcut is 0.
    Makes a highpass filter if highcut is fs/2.
    Apply filter with: `scipy.signal.lfilter(b, 1, data, axis=axis)`
    RH 2021

        Args:
            lowcut (scalar): 
                frequency (in Hz) of low pass band
            highcut (scalar):  
                frequency (in Hz) of high pass band
            num_taps (int): 
                number of taps in the filter. If None, then q is used.
            q (scalar):
                quality factor. \n
                ``num_taps = int(fs / f) * q``. \n
                where f is lowcut if lowcut > 0 else highcut. \n
                If None, then num_taps is used.
            fs (scalar): 
                sample rate (frequency in Hz)
            window (string):
                window to use for the FIR filter
                (see scipy.signal.firwin for all possible inputs)
            plot_pref (bool):
        
        Returns:
            b (ndarray): 
                FIR filter coefficients
    '''
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq

    if q is not None:
        f = lowcut if lowcut > 0 else highcut
        num_taps = int((fs / lowcut) * q)
        num_taps = num_taps + 1 if num_taps % 2 == 0 else num_taps

    if low <= 0:
        ## Make a lowpass filter
        b = scipy.signal.firwin(numtaps=num_taps, cutoff=high, window=window, pass_zero=True)
    elif high >= 1:
        ## Make a highpass filter
        b = scipy.signal.firwin(numtaps=num_taps, cutoff=low, window=window, pass_zero=False)
    else:
        b = scipy.signal.firwin(numtaps=num_taps, cutoff=[low, high], window=window, pass_zero=False)

    if plot_pref:
        plot_digital_filter_response(b=b, fs=fs, worN=100000)
    return b


def plot_digital_filter_response(b, a=None, fs=30, worN=100000, plot_pref=True):
    '''
    plots the frequency response of a digital filter
    RH 2021

        Args:
            b (ndarray): 
                Numerator polynomial coeffs of the IIR filter
            a (ndarray): 
                Denominator polynomials coeffs of the IIR filter
            fs (scalar): 
                sample rate (frequency in Hz)
            worN (int): 
                number of frequencies at which to evaluate the filter
    '''
    w, h = scipy.signal.freqz(b, a, worN=worN) if a is not None else scipy.signal.freqz(b, worN=worN)
    xAxis = (fs * 0.5 / np.pi) * w

    if plot_pref:
        plt.figure()
        plt.plot(xAxis, abs(h))
        plt.xlabel('frequency (Hz)')
        plt.ylabel('frequency response (a.u)')
        plt.xscale('log')

    return xAxis, abs(h)


def mtaper_specgram(
    signal,
    nw=2.5,
    ntapers=None,
    win_len=0.1,
    win_overlap=0.09,
    fs=int(192e3),
    clip=None,
    freq_res_frac=1,
    mode='psd',
    **kwargs
):
    """
    Multi-taper spectrogram.
    From Jeff Markowitz.

        Args:
            signal (array type): 
                Signal.
            nw (float): 
                Time-bandwidth product
            ntapers (int): 
                Number of tapers (None to set to 2 * nw -1)
            win_len (float): 
                Window length in seconds
            win_overlap (float): 
                Window overlap in seconds
            fs (float): 
                Sampling rate in Hz
            clip (2-tuple of floats): 
                Normalize amplitudes to 0-1 using clips (in dB)
            freq_res_frac (float): 
                frequency resolution fraction. 
                generates nfft. If none then nfft=None,
                which makes nfft=win nfft=nperseg=len_samples. 
                else nfft = freq_resolution_frac * round(win_len * fs)
            mode (string): 
                mode of the scipy.signal.spectrogram to use. Can be
                'psd', 'complex', 'magnitude', 'angle', 'phase'
            **kwargs: 
                Additional arguments for scipy.signal.spectrogram
        Returns:
            f (ndarray): 
                Frequency bin centers
            t (ndarray): 
                Time indices
            sxx (ndarray): 
                Spectrogram
    """
    len_samples = np.round(win_len * fs).astype("int")
    if freq_res_frac is None:
        nfft = None
    else:
        nfft = freq_res_frac*len_samples
    if ntapers is None:
        ntapers = int(nw * 2)
    overlap_samples = np.round(win_overlap * fs)
    sequences, r = scipy.signal.windows.dpss(
        len_samples, NW=nw, Kmax=ntapers, sym=False, norm=2, return_ratios=True
    )
    sxx_ls = None
    for sequence, weight in zip(sequences, r):
        f, t, sxx = scipy.signal.spectrogram(
            signal,
            fs=fs,
            window=sequence,
            nperseg=len_samples,
            noverlap=overlap_samples,
            nfft=nfft,
            detrend='constant',
            return_onesided=True, 
            scaling='density', 
            axis=-1, 
            mode=mode,
            **kwargs
        )
        if sxx_ls is None:
            sxx_ls = sxx * weight
        else:
            sxx_ls += np.abs(sxx * weight)
    sxx = sxx_ls / len(sequences)
    if clip is not None:
        sxx = 20 * np.log10(sxx)
        sxx = sxx - clip[0]
        sxx[sxx < 0] = 0
        sxx[sxx > (clip[1] - clip[0])] = clip[1] - clip[0]
        sxx /= clip[1] - clip[0]
    return f, t, sxx


def simple_cwt(
    X,
    freqs_toUse=None, 
    fs=30, 
    wavelet_type='cmor', 
    bwf=None, 
    cf=None, 
    psd_scaling=True, 
    plot_pref=True, 
    axis=-1):
    '''
    performs a simple continuous wavelet transform (cwt) using pywt.cwt
    RH 2021

        Args:
            X (ndarray): 
                data array
            freqs_toUse (1-D ndarray): 
                values of frequencies to perform cwt on
            fs (scalar): sample rate in Hz
            wavelet_type (string): 
                name of wavelet type to use. See pywt.wavelist() for all 
                possible inputs
            bwf (scalar): 
                bandwidth (in units of frequency). Used only if using complex 
                morlet ('cmor')
            cf (scalar): 
                center frequency. Used only if using complex morlet ('cmor')
            axis (int): 
                axis along which to perform cwt
            psd_scaling (bool): 
                preference of whether to scale the output to compute the power 
                spectral density or leave as raw output of pywt.cwt 
        Returns:
            coeff (ndarray): 
                output cwt array (with temporal dimension=='axis').
                A natural way to normalize output is to put it in units of
                'spectral density' = np.abs(coeff**2 / (1/freqs_toUse)[:,None])
                Another nice normalization is
                np.abs(coeff / (1/freqs_toUse)[:,None]**1)**1.5
    '''
    import pywt

    if wavelet_type=='cmor' and bwf is None:
        bwf = 2
    if wavelet_type=='cmor' and cf is None:
        cf = 1
    waveletname = f'{wavelet_type}{bwf}-{cf}'

    if freqs_toUse is None:
        freqs_toUse = np.logspace(np.log(fs/30), np.log(fs/2), 30, base=np.exp(1))
    scales = fs/freqs_toUse


    coeff, freq = pywt.cwt(data=X, 
                        scales=scales, 
                        wavelet=waveletname, 
                        sampling_period=1/fs,
                        axis=axis)

    if psd_scaling:
        coeff = np.abs(coeff**2 / (1/freqs_toUse)[:,None])

    if plot_pref:
        n_ticks = min(len(freq) , 10)
        tick_spacing = len(freq)//n_ticks
        ticks_toUse = np.arange(0,len(freq), tick_spacing)    
        fig, ax = plt.subplots()
        if psd_scaling:
            ax.imshow(coeff, aspect='auto', origin='lower')
        else:
            ax.imshow(np.abs(coeff), aspect='auto', origin='lower')
        ax.set_yticks(ticks_toUse)
        ax.set_yticklabels(np.round(freq[ticks_toUse],2))

    return coeff, freq


@torch.jit.script
def torch_hilbert(x: torch.Tensor, N: Optional[int] = None, dim: int = -1):
    """
    Computes the analytic signal using the Hilbert transform.
    Based on scipy.signal.hilbert
    RH 2022
    
    Args:
        x (nd tensor):
            Signal data. Must be real.
        N (int):
            Number of Fourier components to use.
            If None, then N = x.shape[dim]
        dim (int):
            Dimension along which to do the transformation.
    
    Returns:
        xa (nd tensor):
            Analytic signal of input x along dim
    """
    assert x.is_complex() == False, "x should be real"
    n = x.shape[dim] if N is None else N
    assert n >= 0, "N must be non-negative"

    xf = torch.fft.fft(input=x, n=n, dim=dim)
    m = torch.zeros(n, dtype=xf.dtype, device=xf.device)
    if n % 2 == 0: ## then even
        m[0] = m[n//2] = 1
        m[1:n//2] = 2
    else:
        m[0] = 1 ## then odd
        m[1:(n+1)//2] = 2

    if x.ndim > 1:
        ## pad m with singleton dimensions
        new_shape = [1] * x.ndim
        new_shape[dim] = -1
        m = m.reshape(new_shape)

    return torch.fft.ifft(xf * m, dim=dim)


def hilbert_fourier_domain(xf, axis=-1):
    """
    Computes the Fourier-transform of the analytic signal using the
    Fourier-transformed signal. Input is ``xf`` and output is ``xhf``. Uses the
    approach used in scipy.signal.hilbert.\n
    RH 2024
    
    Args:
        xf (torch.Tensor or np.ndarray):
            Fourier-transformed signal data. Should be complex and have negative
            frequencies. Generally, the output of torch.fft.fft or np.fft.fft.
        axis (int):
            Dimension along which to do the transformation.

    Returns:
        (nd tensor):
            Fourier-transformed analytic signal along axis.
    """
    if isinstance(xf, torch.Tensor):
        is_complex = torch.is_complex
        zeros = functools.partial(torch.zeros, device=xf.device)
    if isinstance(xf, np.ndarray):
        zeros, is_complex = np.zeros, np.iscomplexobj
    
    assert is_complex(xf), "xf should be complex"

    n = xf.shape[axis]

    m = zeros(n, dtype=xf.dtype)
    if n % 2 == 0: ## then even
        m[0] = m[n//2] = 1
        m[1:n//2] = 2
    else:
        m[0] = 1 ## then odd
        m[1:(n+1)//2] = 2

    if xf.ndim > 1:
        ind = [np.newaxis] * xf.ndim
        ind[axis] = slice(None)
        m = m[tuple(ind)]

    return xf * m        
        

def signal_angle_difference(x, y, center=True, window=None, axis=-1):
    """
    Computes the average angle difference between two signals.\n
    Calculated as the mean of the angle difference between the Hilbert
    transforms of the signals.\n
    Bound to the range [-pi, pi] (i.e. -180 to 180 degrees). Negative values
    indicate that x leads y.\n
    RH 2024
    
    Args:
        x (torch.Tensor or np.ndarray):
            Signal data. Must be complex.
        y (torch.Tensor or np.ndarray):
            Signal data. Must be complex.
        center (bool):
            Whether or not to center the signals.
        window (torch.Tensor or np.ndarray):
            If not None, then the signals are multiplied by the window after the
            Hilbert transform.
        dim (int):
            Dimension along which to do the transformation.
    
    Returns:
        angle_diff (nd tensor):
            Angle difference between x and y along dim
    """
    if isinstance(x, torch.Tensor) and isinstance(y, torch.Tensor):
        mean, circmean, hilbert = (functools.partial(fn, axis=axis) for fn in (torch.mean, torch_helpers.circmean, torch_hilbert))
        angle, pi = torch.angle, torch.pi
    elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        mean, circmean, hilbert = (functools.partial(fn, axis=axis) for fn in (np.mean, scipy.stats.circmean, scipy.signal.hilbert))
        angle, pi = np.angle, np.pi
    else:
        raise ValueError("x and y must be torch tensors or numpy arrays")
    
    fn_win = lambda x: x * window if window is not None else x
    
    if center:
        x, y = x - mean(x, keepdims=True), y - mean(y, keepdims=True)
    
    out = circmean((angle(fn_win(hilbert(x))) - angle(fn_win(hilbert(y)))))

    if out.ndim == 0:
        out = out - 2*pi if out > pi else out
    else:
        out[out > pi] -= 2*pi
    out = out[()]
    return out


@misc.wrapper_flexible_args(['dim', 'axis'])
def time_domain_reversal_in_fourier_domain(xf, axis=-1):
    """
    Reverses the Fourier-transformed input signal by effectively applying a
    time-domain reversal while remaining in the Fourier domain. This is done by
    applying the inverse shift operator. \n
    RH 2024

    Args:
        xf (torch.Tensor or np.ndarray):
            Fourier-transformed signal data. Should be complex and have negative
            frequencies. Generally, the output of torch.fft.fft or np.fft.fft.
        axis (int):
            Dimension along which to do the transformation.

    Returns:
        (nd tensor):
            Time-domain signal reversed along axis
    """
    if isinstance(xf, torch.Tensor):
        assert torch.is_complex(xf), "xf should be complex"
        arange = functools.partial(torch.arange, device=xf.device, dtype=torch_helpers.dtype_to_real(xf.dtype))
        exp, pi, conj = torch.exp, torch.pi, torch.conj
        resolution = lambda x: x.resolve_conj().resolve_neg()
    elif isinstance(xf, np.ndarray):
        arange = np.arange
        exp, pi, conj = np.exp, np.pi, np.conj
        resolution = lambda x: x
    
    ## Make the inverse shift operator
    ### The 'iso' is one period of a complex sinuoid. 
    ### It basically changes the ratio of real to imaginary parts in xf.
    n = xf.shape[axis]
    k = arange(n)
    w = exp(-2j * pi * k / n)
    ## Apply the inverse shift operator
    return resolution(conj(xf * w))


def filtfilt_simple_fft(
    x: Union[torch.Tensor, np.ndarray],
    kernel: Union[torch.Tensor, np.ndarray],
    fast_len: bool = True,
):
    """
    Applies a zero-phase filter to the input signal using the FFT method along
    the last dimension.\n
    Calculated as ``ifft(fft(x) * fft(kernel) * fft(kernel_flipped))``.\n
    This implementation is very fast and is suitable when using long kernels.\n
    NOTE: This is a simple implementation and does not handle edge effects.
    scipy.signal.filtfilt is recommended if speed is not a concern and/or if
    kernel length is similar in length to x.\n
    RH 2024

    Args:
        x (torch.Tensor or np.ndarray):
            Signal data. Convolution is done along the last dimension.
        kernel (torch.Tensor or np.ndarray):
            Filter kernel. Convolution is done along the last dimension. \n
            If not 1D, then shape should be broadcastable with x.
        fast_len (bool):
            Whether to use the fast length method.

    Returns:
        (nd tensor):
            Filtered signal
    """
    assert isinstance(x, torch.Tensor) or isinstance(x, np.ndarray), "x must be a torch tensor or numpy array"

    if isinstance(x, torch.Tensor) and isinstance(kernel, torch.Tensor):
        use_real = (torch.is_complex(x) == False) and (torch.is_complex(kernel) == False)
        fft, ifft = (functools.partial(fn, dim=-1) for fn in ((torch.fft.rfft, torch.fft.irfft) if use_real else (torch.fft.fft, torch.fft.ifft)))
        flip = functools.partial(torch.flip, dims=(-1,))
    elif isinstance(x, np.ndarray) and isinstance(kernel, np.ndarray):
        use_real = (np.iscomplexobj(x) == False) and (np.iscomplexobj(kernel) == False)
        fft, ifft = (functools.partial(fn, axis=-1) for fn in ((np.fft.rfft, np.fft.irfft) if use_real else (np.fft.fft, np.fft.ifft)))
        flip = functools.partial(np.flip, axis=-1)
    else:
        raise ValueError("x and kernel must be torch tensors or numpy arrays")
    
    f_flip = functools.partial(time_domain_reversal_in_fourier_domain, axis=-1)

    n = x.shape[-1] + kernel.shape[-1] - 1
    n = timeSeries.next_fast_len(n) if fast_len else n

    out = fft(x, n=n)  ## x_fft
    kernel_fft = fft(flip(kernel), n=n)
    out = out * kernel_fft  ## xk_fft_1
    out = out * f_flip(kernel_fft)  ## xk_fft_2
    out = ifft(out, n=n)  ## xk
    out = torch_helpers.slice_along_dim(
        X=out,
        dim=-1,
        idx=slice(0, x.shape[-1]),
    )
    out = out.real if use_real else out
    return out


@torch.jit.script
def _helper_time_domain_reversal_in_fourier_domain(xf: torch.Tensor, axis: int = -1):
    """
    Reverses the Fourier-transformed input signal by effectively applying a
    time-domain reversal while remaining in the Fourier domain. This is done by
    applying the inverse shift operator. \n
    RH 2024

    Args:
        xf (torch.Tensor or np.ndarray):
            Fourier-transformed signal data. Should be complex and have negative
            frequencies. Generally, the output of torch.fft.fft or np.fft.fft.
        axis (int):
            Dimension along which to do the transformation.

    Returns:
        (nd tensor):
            Time-domain signal reversed along axis
    """
    assert torch.is_complex(xf), "xf should be complex"
    
    ## Make the inverse shift operator
    ### The 'iso' is one period of a complex sinuoid. 
    ### It basically changes the ratio of real to imaginary parts in xf.
    n = xf.shape[axis]
    k = torch.arange(n, device=xf.device, dtype=torch_helpers.dtype_to_real(xf.dtype))
    w = torch.exp(-2j * torch.pi * k / n)
    ## Apply the inverse shift operator
    return torch.conj(xf * w).resolve_conj().resolve_neg()
@torch.jit.script
def torch_filtfilt_simple_fft(
    x: torch.Tensor,
    kernel: torch.Tensor,
    fast_len: bool = True,
):
    """
    Exactly the same as ``filtfilt_simple_fft`` but works with torch.jit.script.
    Applies a zero-phase filter to the input signal using the FFT method along
    the last dimension.\n
    Calculated as ``ifft(fft(x) * fft(kernel) * fft(kernel_flipped))``.\n
    This implementation is very fast and is suitable when using long kernels.\n
    NOTE: This is a simple implementation and does not handle edge effects.
    scipy.signal.filtfilt is recommended if speed is not a concern and/or if
    kernel length is similar in length to x.\n
    RH 2024

    Args:
        x (torch.Tensor or np.ndarray):
            Signal data. Convolution is done along the last dimension.
        kernel (torch.Tensor or np.ndarray):
            Filter kernel. Convolution is done along the last dimension. \n
            If not 1D, then shape should be broadcastable with x.
        fast_len (bool):
            Whether to use the fast length method.

    Returns:
        (nd tensor):
            Filtered signal
    """
    use_real = (torch.is_complex(x) == False) and (torch.is_complex(kernel) == False)

    n = x.shape[-1] + kernel.shape[-1] - 1
    n = timeSeries.next_fast_len(n) if fast_len else n

    out = torch.fft.fft(x, n=n, dim=-1)  ## x_fft
    kernel_fft = torch.fft.fft(torch.flip(kernel, dims=(-1,)), n=n, dim=-1)
    out = out * kernel_fft  ## xk_fft_1
    out = out * _helper_time_domain_reversal_in_fourier_domain(kernel_fft, axis=-1)  ## xk_fft_2
    out = torch.fft.ifft(out, n=n, dim=-1)  ## xk
    # out = torch_helpers.slice_along_dim(
    #     X=out,
    #     dim=-1,
    #     idx=slice(0, x.shape[-1]),
    # )
    out = out[..., :x.shape[-1]]
    out = out.real if use_real else out
    return out

def torch_coherence(
    x: torch.Tensor,
    y: torch.Tensor,
    fs: float = 1.0,
    window: str = 'hann',
    nperseg: Optional[int] = None,
    noverlap: Optional[int] = None,
    nfft: Optional[int] = None,
    detrend: str = 'constant',
    axis: int = -1,
    batch_size: Optional[int] = None,
    pad_last_segment: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Computes the magnitude-squared coherence between two signals using a PyTorch
    implementation. This function gives identical results to the
    scipy.signal.coherence, but its much faster. A ``batch_size`` option allows
    for large arrays to be processed on GPU.\n    
    RH 2024
    
    Args:
        x (torch.Tensor): 
            First input signal.
        y (torch.Tensor): 
            Second input signal.
        fs (float): 
            Sampling frequency of the input signal. (Default is 1.0)
        window (str): 
            Type of window to apply. Supported window types are the same as
            `scipy.signal.get_window`. (Default is 'hann')
        nperseg (Optional[int]): 
            Length of each segment. (Default is ``None``, which uses ``len(x) //
            8``)
        noverlap (Optional[int]): 
            Number of points to overlap between segments. (Default is ``None``,
            which uses ``nperseg // 2``)
        nfft (Optional[int]): 
            Number of points in the FFT used for each segment. (Default is
            ``None``, which sets it equal to `nperseg`)
        detrend (str): 
            Specifies how to detrend each segment. Supported values are
            'constant' or 'linear'. (Default is 'constant')
        axis (int): 
            Axis along which the coherence is calculated. (Default is -1)
        batch_size (Optional[int]):
            Number of segments to process at once. Used to reduce memory usage.
            If None, then all segments are processed at once. Note that
            ``num_segments = (x.shape[axis] - nperseg) // (nperseg - noverlap) +
            1``. (Default is None)
        pad_last_segment (bool):
            Whether to pad the last segment with a value defined by this
            argument. If None, then no padding is done. (Default is None)

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: 
            freqs (torch.Tensor): 
                Frequencies for which the coherence is computed. Shape is
                ``nfft // 2 + 1``.
            coherence (torch.Tensor): 
                Magnitude-squared coherence values. Shape is x broadcasted with
                y and the specified axis set to ``nfft // 2 + 1``.

    Example:
        .. highlight:: python
        .. code-block:: python

            x = torch.randn(1024)
            y = torch.randn(1024)
            freqs, coherence = torch_coherence(x, y, fs=256)
    """
    ## Convert axis to positive
    axis = axis % len(x.shape)

    ## Checks
    ### Tensor checks
    assert isinstance(x, torch.Tensor), "x should be a torch tensor"
    assert isinstance(y, torch.Tensor), "y should be a torch tensor"
    ### Dims should either be the same or one of them should be 1
    if not (x.shape == y.shape):
        assert all([(x.shape[ii] == y.shape[ii]) or (1 in [x.shape[ii], y.shape[ii]]) for ii in range(len(x.shape))]), f"x and y should have the same shape or one of them should have shape 1 at each dimension. Found x.shape={x.shape} and y.shape={y.shape}"

    if nperseg is None:
        nperseg = len(x) // 8

    if noverlap is None:
        noverlap = nperseg // 2

    if nfft is None:
        nfft = nperseg

    if window is not None:
        window = scipy.signal.get_window(window, nperseg)
        window = torch.tensor(window, dtype=x.dtype, device=x.device)

    ## args should be greater than nfft
    assert all([arg.shape[axis] >= nfft for arg in (x, y)]), f"Signal length along axis should be greater than nfft. Found x.shape={x.shape} and y.shape={y.shape} and nfft={nfft}"

    ## Detrend the signals
    def detrend_constant(y, axis):
        y = y - torch.mean(y, axis=axis, keepdim=True)
        return y

    if detrend == 'linear':
        X_linearDetrendPrep = torch.ones(nperseg, 2, dtype=y.dtype, device=y.device)
        X_linearDetrendPrep[:, 1] = torch.arange(nperseg, dtype=y.dtype, device=y.device)
    def detrend_linear(y, axis):
        """
        Uses least squares approach to remove linear trend.
        """
        ## Move axis to end
        y = y.moveaxis(axis, -1)[..., None]
        ## Prepare the design matrix
        X = X_linearDetrendPrep[([None] * (len(y.shape) - 2))]
        ## Compute the coefficients
        # beta = torch.linalg.lstsq(X, y)[0] 
        ### Use closed form solution for least squares
        beta = torch.linalg.inv(X.transpose(-1, -2) @ X) @ X.transpose(-1, -2) @ y
        ## Remove the trend
        y = y - opt_einsum.contract('...ij, ...jk -> ...ik', X, beta)
        y = y[..., 0]
        ## Move axis back to original position (argsort y_dims_to)
        y = y.moveaxis(-1, axis)
        return y

    if detrend == 'constant':
        fn_detrend = detrend_constant
    elif detrend == 'linear':
        fn_detrend = detrend_linear
    else:
        raise ValueError(f"detrend must be 'constant' or 'linear'. Found {detrend}")
    
    ## Initialize the coherence arrays
    ### Get broadcasted dimensions: max(x, y) at each dimension, and nfft at axis
    x_shape = list(x.shape)
    y_shape = list(y.shape)
    out_shape = [max(x_shape[i], y_shape[i]) for i in range(len(x_shape))]
    out_shape[axis] = nfft // 2 + 1  ## rfft returns only non-negative frequencies (0 to fs/2 inclusive)
    
    ## Initialize sums for Welch's method
    ### Prepare complex dtype
    dtype_complex = torch_helpers.dtype_to_complex(x.dtype)
    f_cross = torch.zeros(out_shape, dtype=dtype_complex, device=x.device)
    psd1 = torch.zeros(out_shape, dtype=x.dtype, device=x.device)
    psd2 = torch.zeros(out_shape, dtype=x.dtype, device=x.device)

    ## Prepare batch generator
    num_segments = (x.shape[axis] - nperseg) // (nperseg - noverlap) + 1
    batch_size = max(num_segments, 1) if batch_size is None else batch_size
    x_batches, y_batches = (indexing.batched_unfold(
        var, 
        dimension=axis, 
        size=nperseg, 
        step=nperseg - noverlap, 
        batch_size=batch_size,
        pad_value=pad_last_segment,
    ) for var in (x, y))
    
    process_segment = lambda x: torch.fft.rfft(fn_detrend(x, axis=-1) * window, n=nfft, dim=-1)  ## Note the broadcasting of 1-D window with last dimension of x
    ## Perform Welch's averaging of FFT segments
    for segs_x, segs_y in zip(x_batches, y_batches):
        segs_x = process_segment(segs_x)
        segs_y = process_segment(segs_y)
        f_cross += (torch.sum(torch.conj(segs_x) * segs_y, dim=axis).moveaxis(-1, axis)) / num_segments
        psd1 += (torch.sum((torch.conj(segs_x) * segs_x).real, dim=axis).moveaxis(-1, axis)) / num_segments
        psd2 += (torch.sum((torch.conj(segs_y) * segs_y).real, dim=axis).moveaxis(-1, axis)) / num_segments

    ## Compute coherence
    coherence = torch.abs(f_cross) ** 2 / (psd1 * psd2)

    ## Generate frequency axis
    freqs = torch.fft.rfftfreq(nfft, d=1 / fs)

    ## Take the positive part of the frequency spectrum
    ### NOTE: This is not necessary as the coherence is symmetric (always odd and real)
    # pos_mask = freqs >= 0
    # ### slice along axis
    # freqs = freqs[pos_mask]
    # coherence = torch_helpers.slice_along_dim(coherence, axis=axis, idx=pos_mask)
    
    return freqs, coherence


def spectrogram_magnitude_normalization(S: torch.Tensor, k: float = 0.05):
    """
    Normalize spectrogram by dividing by the total power across all frequencies
    at each time point. ``mag = mag / ((k * mean_mag_at_each_time) + (1-k))``.
    This formula differs slightly from the forumla often used in audio
    engineering in that it is a linear scaling, and does not apply only to data
    above a certain threshold.

    Args:
        spectrogram (torch.Tensor):
            A single spectrogram
            Spectrogram should be of shape: (n_points, n_freq_bins, n_samples)  
        k (float):
            Weighting factor for the normalization. 0 means no normalization. 1
            means every time point has the same power when summed across all
            frequencies.

    Returns:
        spectrogram (tuple or torch.Tensor):
            A single normalized spectrogram of same shape as input.
    """
    if k == 0:
        return S
    
    ## Check inputs
    if isinstance(S, torch.Tensor):
        mean, sum, polar, is_complex, abs, angle  = torch.mean, torch.sum, torch.polar, torch.is_complex, torch.abs, torch.angle
    elif isinstance(S, np.ndarray):
        mean, sum, is_complex, abs, angle = np.mean, np.sum, np.iscomplexobj, np.abs, np.angle
        polar = lambda mag, phase: mag * np.exp(1j * phase)

    assert S.ndim == 3, "Spectrogram should be of shape: (n_points, n_freq_bins, n_samples)"

    if is_complex(S) == False:
        s_mean = mean(sum(S, dim=1, keepdims=True), dim=0, keepdims=True)  ## Mean of the summed power across all frequencies and points. Shape (n_samples,)
        s_norm =  S / ((k * s_mean) + (1-k))  ## Normalize the spectrogram by the mean power across all frequencies and points. Shape (n_points, n_freq_bins, n_samples)
    
    elif is_complex(S) == True:
        s_mag = abs(S)
        s_phase = angle(S)
        s_mean = mean(sum(s_mag, dim=1, keepdims=True), dim=0, keepdims=True)  ## Mean of the summed power across all frequencies and points. Shape (n_samples,)
        s_mag = s_mag / ((k * s_mean) + (1-k))  ## Normalize the spectrogram by the mean power across all frequencies and points. Shape (n_points, n_freq_bins, n_samples)
        s_norm = polar(s_mag, s_phase)
    
    return s_norm



def ppc(phases, axis=-1):
    """
    Computes the pairwise phase consistency (PPC0) for a (set of) vector of
    phases. Based on Vinck et al. 2010, and the implementation in the FieldTrip
    toolbox:
    https://github.com/fieldtrip/fieldtrip/blob/d7403c6a6e8765b679ba8accc69f69a282fce6cf/contrib/spike/ft_spiketriggeredspectrum_stat.m#L442
    RH 2024

    Args:
        phases (np.ndarray): 
            Vector of phases in radians. Bound to the range [-pi, pi].
        axis (int):
            Axis along which to compute the pairwise phase consistency.

    Returns:
        float: 
            Pairwise phase consistency of the phases.
    """
    if isinstance(phases, torch.Tensor):
        sin, cos, abs, sum = torch.sin, torch.cos, torch.abs, torch.sum
    elif isinstance(phases, np.ndarray):
        sin, cos, abs, sum = np.sin, np.cos, np.abs, np.sum

    N = phases.shape[axis]
    if N < 2:
        raise ValueError("The input vector must contain at least two phase values.")

    # Compute pairwise phase consistency
    sinSum = abs(sum(sin(phases), axis=axis))
    cosSum = sum(cos(phases), axis=axis)
    return ((cosSum**2 + sinSum**2) - N) / (N * (N - 1))


@torch.jit.script
def torch_ppc(phases: torch.Tensor, axis: int = -1):
    """
    Exactly the same as ``ppc`` but works with torch.jit.script.
    Computes the pairwise phase consistency (PPC0) for a (set of) vector of
    phases. Based on Vinck et al. 2010, and the implementation in the FieldTrip
    toolbox:
    https://github.com/fieldtrip/fieldtrip/blob/d7403c6a6e8765b679ba8accc69f69a282fce6cf/contrib/spike/ft_spiketriggeredspectrum_stat.m#L442
    RH 2024

    Args:
        phases (np.ndarray): 
            Vector of phases in radians. Bound to the range [-pi, pi].
        axis (int):
            Axis along which to compute the pairwise phase consistency.

    Returns:
        float: 
            Pairwise phase consistency of the phases.
    """
    N = phases.shape[axis]
    if N < 2:
        raise ValueError("The input vector must contain at least two phase values.")

    # Compute pairwise phase consistency
    sinSum = torch.abs(torch.sum(torch.sin(phases), dim=axis))
    cosSum = torch.sum(torch.cos(phases), dim=axis)
    return ((cosSum**2 + sinSum**2) - N) / (N * (N - 1))


def ppc_windowed(phases, window, axis=-1):
    """
    Computes the pairwise phase consistency (PPC0) for a (set of) vector of
    phases using a windowed approach.
    RH 2024

    Args:
        phases (np.ndarray): 
            Matrix of phases in radians. Bound to the range [-pi, pi].
        window (int): 
            Vector or matrix of same length as phases along axis. Window to
            multiply by the sin and cos components.
            If 1-D: Shape should be (n_points,)
            If N-D: Shape must be broadcastable with phases. window.shape[axis]
            must be equal to phases.shape[axis]
        axis (int):
            Axis along which to compute the pairwise phase consistency.

    Returns:
        float: 
            Pairwise phase consistency of the phases.
    """
    if isinstance(phases, torch.Tensor):
        sin, cos, abs, sum = torch.sin, torch.cos, torch.abs, torch.sum
    elif isinstance(phases, np.ndarray):
        sin, cos, abs, sum = np.sin, np.cos, np.abs, np.sum

    N = phases.shape[axis]
    if N < 2:
        raise ValueError("The input vector must contain at least two phase values.")

    if window.ndim == 1:
        assert window.shape[0] == N, "Window should have the same length as the input phases."
    else:
        assert window.shape[axis] == N, "Window should have the same length as the input phases."

    # Compute pairwise phase consistency
    sinSum = abs(sum(sin(phases) * window, axis=axis))
    cosSum = sum(cos(phases) * window, axis=axis)
    return ((cosSum**2 + sinSum**2) - N) / (N * (N - 1))


def generate_multiphasic_sinewave(
    n_samples: int = 10000,
    n_periods: float = 1.0,
    n_waves: int = 3,
    return_x: bool = False,
    return_phases: bool = False,
):
    """
    Generates a multiphasic sine wave.
    RH 2024

    Args:
        n_samples (int): 
            Number of samples to generate.
        n_periods (float): 
            Number of periods in the sine wave.
        n_waves (int): 
            Number of sine waves to generate.
        return_x (bool): 
            If ``True``, returns the x-values along with the waves.
        return_phases (bool): 
            If ``True``, returns the phases along with the waves.

    Returns:
        (tuple): 
            Depending on the `return_x` and `return_phases` parameters, the
            function returns some combination of the following: \n
                * waves (np.ndarray): The generated sine waves.
                * x (np.ndarray): The x-values.
                * phases (np.ndarray): The phases of the sine waves.
    """
    x = np.linspace(0, n_periods * np.pi*2, n_samples)

    phases = np.stack([
        x - ii * np.pi * (2 / n_waves) for ii in range(n_waves)
    ])
    waves = np.cos(phases)
    
    if return_x and return_phases:
        return waves, x, phases
    elif return_x:
        return waves, x
    elif return_phases:
        return waves, phases
    else:
        return waves